# -*- encoding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import time
import os
import socket
import re
from rbfZOS.py3270 import Emulator
from robot.api import logger
from robot.api.deco import keyword
from robot.libraries.BuiltIn import BuiltIn
from robot.utils import Matcher


class x3270(object):
    def __init__(self, visible=True, timeout='30', wait_time='0.5', wait_time_after_write='0', img_folder='.'):
        """You can change to hide the emulator screen set the argument visible=${False}
           
           For change the wait_time see `Change Wait Time`, to change the img_folder
           see the `Set Screenshot Folder` and to change the timeout see the `Change Timeout` keyword.
        """
        self.lu = None
        self.host = None
        self.port = None
        self.credential = None
        self.imgfolder = img_folder
        # Try Catch to run in Pycharm, and make a documentation in libdoc with no error
        try:
            self.output_folder = BuiltIn().get_variable_value('${OUTPUT DIR}')
        except Exception as e:
            if hasattr(e, 'message') and e.message == 'Cannot access execution context': self.output_folder = os.getcwd()
#            else: raise AssertionError(e)
        self.wait = float(wait_time)
        self.wait_write = float(wait_time_after_write)
        self.timeout = int(timeout)
        self.mf = Emulator(visible=bool(visible), timeout=int(timeout))

    @keyword(name=None, tags=("Settings",))
    def change_timeout(self, seconds):
        """*Change the timeout for connection in seconds.*
        """
        self.timeout = float(seconds)

    @keyword(name=None, tags=("Settings",))
    def open_connection(self, host, LU=None, port=23):
        """*Create a connection with IBM3270 mainframe with the default port 23.*         
        
        *Arguments*
        
        ``host``: Provide the hostname as IP address to make a connection with mainframe
        
        ``LU``: Optional. Provide the Logical Unit Name
        
        ``port``: Optional. Default port is 23.
        
        
        *Examples*
        
            | `Open Connection` | 10.3.51.71 |
            | `Open Connection` | 10.3.51.71 | LU=LUname |
            | `Open Connection` | 10.3.51.71 | port=992 |
        """
        self.host = host
        self.lu = LU
        self.port = port
        if self.lu:
            self.credential = "%s@%s:%s" % (self.lu, self.host, self.port)
        else:
            self.credential = self.host
        self.mf.connect(self.credential)

    @keyword(name=None, tags=("Settings",))
    def close_connection(self):
        """*Disconnect from the host.*
        """
        try:
            self.mf.terminate()
        except socket.error:
            pass

    @keyword(name=None, tags=("Settings",))
    def change_wait_time(self, wait_time):
        """*To give time for the mainframe screen to be "drawn" and receive the next commands.* 
        To change the default value of 0.5, use this keyword passing the time in seconds.
        
        *Arguments*
        
        ``wait_time``: Provide the wait time so that a sleep applies after the following keywords `Execute Command`, `Send Enter`, `Send PF`, `Write`, `Write in position`
        
        *Examples*
        
            | `Change Wait Time` | 0.5 |
            | `Change Wait Time` | 2 |
        """
        self.wait = float(wait_time)

    @keyword(name=None, tags=("Settings",))
    def change_wait_time_after_write(self, wait_time_after_write):
        """*To give the user time to see what is happening inside the mainframe.*
        To change the default value of 0 seconds, use this keyword passing the time in seconds. 
        Note: This keyword is useful for debug purpose
        
        *Arguments*
        
        ``wait_time_after_write``: Provide the wait time so that a sleep applies after the string is sent in the following keywords `Write`, `Write Bare`, `Write in position`, `Write Bare in position`       

        Examples:
            | Change Wait Time After Write | 0.5 |
            | Change Wait Time After Write | 2 |
        """
        self.wait_write = float(wait_time_after_write)

    @keyword(name=None, tags=("Settings",))
    def read(self, ypos, xpos, length):
        """*Read a string of "length" at screen co-ordinates "ypos"/"xpos".*
           
           
           *Arguments*
           
           ``ypos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal. 
           
           ``xpos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal.
           
           ``length``: Numeric Unit. Length of the string
           
           
           *Examples*
           
               | ${value} | Read | 8 | 10 | 15 | #Read a string in the position y=8 / x=10 of a length 15 |
        """
        self._check_limits(ypos, xpos)
        # Checks if the user has passed a length that will be larger than the x limit of the screen.
        if int(xpos) + int(length) > (80+1):
            raise Exception('You have exceeded the x-axis limit of the mainframe screen')
        string = self.mf.string_get(int(ypos), int(xpos), int(length))
        return str(string)

    @keyword(name=None, tags=("Settings",))
    def execute_command(self, cmd):
        """Execute x3270 command.*
           
           *Arguments*
           
           ``cmd``: [http://x3270.bgp.nu/wc3270-man.html#Actions|x3270 command] 
           
           
           *Examples*
           
               | `Execute Command` | Enter |
               | `Execute Command` | Home |
               | `Execute Command` | Tab |
               | `Execute Command` | PF(1) |
        """
        self.mf.exec_command((str(cmd)).encode("utf-8"))
        time.sleep(self.wait)

    @keyword(name=None, tags=("Settings",))
    def set_screenshot_folder(self, path):
        """*Set a folder to keep the html files generated by the `Take Screenshot` keyword.*
           
           *Arguments*
           
           ``path``: location of folder with absolute path 
           
           *Examples*
           
               | Set Screenshot Folder | /path/to/folder |
        """
        if os.path.exists(os.path.normpath(os.path.join(self.output_folder, path))):
            self.imgfolder = path
        else:
            logger.error('Given screenshots path "%s" does not exist' % path)
            logger.warn('Screenshots will be saved in "%s"' % self.imgfolder)

    @keyword(name=None, tags=("Settings",))
    def take_screenshot(self, height='410', width='670'):
        """*Generate a screenshot of the IBM 3270 Mainframe in a html format.* 
        The default folder is the log folder of RobotFramework, if you want change see the `Set Screenshot Folder`.

           The Screenshot is printed in a iframe log, with the values of height=410 and width=670, you
           can change this values passing them from the keyword. 

           *Arguments*
           
           ``height``: Default is 410.
           
           ``width``: Default is 670.   
           
           *Examples*
           
               | `Take Screenshot` |
               | `Take Screenshot` | height=500 | width=700 |
        """
        filename_prefix = 'screenshot'
        extension = 'html'
        filename_sufix = str(int(round(time.time() * 1000)))
        filepath = os.path.join(self.imgfolder, '%s_%s.%s' % (filename_prefix, filename_sufix, extension))
        self.mf.save_screen(os.path.join(self.output_folder, filepath))
        logger.write('<iframe src="%s" height="%s" width="%s"></iframe>' % (filepath.replace("\\", "/"), height, width),
                     level='INFO', html=True)

    @keyword(name=None, tags=("Settings",))
    def wait_field_detected(self):
        """*Wait until the screen is ready, the cursor has been positioned on a modifiable field, and the keyboard is unlocked.*
        Sometimes the server will "unlock" the keyboard but the screen
        will not yet be ready.  In that case, an attempt to read or write to the
        screen will result in a 'E' keyboard status because we tried to read from
        a screen that is not yet ready.
        Using this method tells the client to wait until a field is
        detected and the cursor has been positioned on it.
        """
        self.mf.wait_for_field()

    @keyword(name=None, tags=("Settings",))
    def delete_char(self, ypos=None, xpos=None):
        """*Delete character under cursor. To delete character in another position, pass the co-ordinates "ypos"/"xpos".*
                     
           
           *Arguments*
           
           ``ypos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal. 
           
           ``xpos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal.   
           
           *Examples*
           
               | `Delete Char` |
               | `Delete Char` | ypos=9 | xpos=25 |
        """
        if ypos is not None and xpos is not None:
            self.mf.move_to(int(ypos), int(xpos))
        self.mf.exec_command(b'Delete')

    @keyword(name=None, tags=("Settings",))
    def delete_field(self, ypos=None, xpos=None):
        """Delete contents in field at current cursor location and positions cursor at beginning of field. To delete a field in another position,  pass the coordinates "ypos"/"xpos".* of any part of the field.
                      
           *Arguments*
           
           ``ypos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal. 
           
           ``xpos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal.   
           
           
           *Examples*
           
               | `Delete Field` |
               | `Delete Field` | ypos=12 | xpos=6 |
        """
        if ypos is not None and xpos is not None:
            self.mf.move_to(int(ypos), int(xpos))
        self.mf.exec_command(b'DeleteField')

    @keyword(name=None, tags=("Settings",))
    def send_enter(self):
        """Send Enter to the screen.*
        """
        self.mf.send_enter()
        time.sleep(self.wait)

    @keyword(name=None, tags=("Settings",))
    def move_next_field(self):
        """*Move cursor to the next input field. Equivalent to pressing the Tab key.*
        """
        self.mf.exec_command(b'Tab')

    @keyword(name=None, tags=("Settings",))
    def move_previous_field(self):
        """*Move the cursor to the previous input field. Equivalent to pressing the Shift+Tab keys.*
        """
        self.mf.exec_command(b'BackTab')

    @keyword(name=None, tags=("Settings",))
    def send_PF(self, PF):
        """*Send a Program Function to the screen.*
        
        *Examples*
        
               | `Send PF` | 3 |
        """
        self.mf.exec_command(('PF('+str(PF)+')').encode("utf-8"))
        time.sleep(self.wait)

    @keyword(name=None, tags=("Settings",))
    def write(self, txt):
        """*Send a string to the screen at the current cursor location. Enter is executed by default with this keyword.*
           
           *Arguments*
           
           ``txt``: String to be passed 
           
           
           *Examples*
           
               | `Write` | something |
        """
        self._write(txt, enter='1')

    @keyword(name=None, tags=("Settings",))
    def write_bare(self, txt):
        """*Send only the string to the screen at the current cursor location.*
           
           *Arguments*
           
           ``txt``: String to be passed 
           
           *Examples*
           
               | `Write Bare` | something |
        """
        self._write(txt)

    @keyword(name=None, tags=("Settings",))
    def write_in_position(self, txt, ypos, xpos):
        """*Send a string to the screen at screen co-ordinates "ypos"/"xpos". Enter is executed by default with this keyword.*
                      
           *Arguments*
           
           ``txt``: String to be passed 
           
           ``ypos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal. 
           
           ``xpos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal.  
           
           *Examples*
           
               | `Write in Position` | something | 9 | 11 |
        """
        self._write(txt, ypos=ypos, xpos=xpos, enter='1')

    @keyword(name=None, tags=("Settings",))
    def write_bare_in_position(self, txt, ypos, xpos):
        """*Send only the string to the screen at screen co-ordinates "ypos"/"xpos".*
           
           *Arguments*
           
           ``txt``: String to be passed 
           
           ``ypos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal. 
           
           ``xpos``: Numeric Unit. Co-ordinate is 1 based, listed in the status area of the terminal.  
           
           
           *Examples*
           
               | `Write Bare in Position` | something | 9 | 11 |
        """
        self._write(txt, ypos=ypos, xpos=xpos)

    def _write(self, txt, ypos=None, xpos=None, enter='0'):
        txt = txt.encode('utf-8')
        if ypos is not None and xpos is not None:
            self._check_limits(int(ypos), int(xpos))
            self.mf.move_to(int(ypos), int(xpos))
        if not isinstance(txt, (list, tuple)): txt = [txt]
        [self.mf.send_string(el) for el in txt if el != '']
        time.sleep(self.wait_write)
        for i in range(int(enter)):
            self.mf.send_enter()
            time.sleep(self.wait)

    @keyword(name=None, tags=("Settings",))
    def wait_until_string(self, txt, timeout=5):
        """*Wait until a string exists on the mainframe screen to perform the next step.* 
        If the string not appear on 5 seconds (default timeout) the keyword will raise a exception. Default timeout can be overridden by passing desired value to ``timeout``
           
           
           *Arguments*
           
           ``txt``: String to be passed 
           
           ``timeout``: Default is 5 seconds. Can be overridden by passing timeout in seconds.
           
           
           *Examples*
           
               | `Wait Until String` | something |
               | `Wait Until String` | something | timeout=10 |
        """
        max_time = time.ctime(int(time.time())+int(timeout))
        while time.ctime(int(time.time())) < max_time:
            result = self._search_string(str(txt))
            if result:
                return txt
        raise Exception('String "' + txt + '" not found in ' + str(timeout) + ' seconds')

    @keyword(name=None, tags=("Settings",))
    def _search_string(self, string, ignore_case=False):
        """*Search if a string exists on the mainframe screen and return True or False.*
        
        *Arguments*
           
           ``string``: String to be passed 
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True 
        
         *Examples*
           
               | `Wait Until String` | something |        
        """
        def __read_screen(string, ignore_case):
            for ypos in range(24):
                line = self.mf.string_get(ypos+1, 1, 80)
                if ignore_case: line = line.lower()
                if string in line:
                    logger.debug('String: ' + string + 'found on mainframe screen')
                    return True
            return False
        status = __read_screen(string, ignore_case)
        return status

    @keyword(name=None, tags=("Settings",))
    def page_should_contain_string(self, txt, ignore_case=False, error_message=None):
        """*Search if a given string exists on the mainframe screen. Raised exception can be overridden with error_message*
           
           *Arguments*
           
           ``txt``: String to be searched 
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True 
           
           ``error_message``: Override the default message with a new value.
            
           
           *Examples*
           
               | `Page Should Contain String` | something |
               | `Page Should Contain String` | someTHING | ignore_case=${True} |
               | `Page Should Contain String` | something | error_message=New error message |
        """
        message = 'The string "' + txt + '" was not found'
        if error_message: message = error_message
        if ignore_case: txt = str(txt).lower()
        result = self._search_string(txt, ignore_case)
        if not result: raise Exception(message)
        logger.info('The string "' + txt + '" was found')

    @keyword(name=None, tags=("Settings",))
    def page_should_not_contain_string(self, txt, ignore_case=False, error_message=None):
        """*Search if a given string NOT exists on the mainframe screen. Raised exception can be overridden with error_message*
           
           *Arguments*
           
           ``txt``: String to be searched 
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True 
           
           ``error_message``: Override the default message with a new value.
           
           
           *Examples*
           
               | `Page Should Not Contain String` | something |
               | `Page Should Not Contain String` | someTHING | ignore_case=${True} |
               | `Page Should Not Contain String` | something | error_message=New error message |
        """
        message = 'The string "' + txt + '" was found'
        if error_message: message = error_message
        if ignore_case: txt = str(txt).lower()
        result = self._search_string(txt, ignore_case)
        if result: raise Exception(message)

    @keyword(name=None, tags=("Settings",))
    def page_should_contain_any_string(self, list_string, ignore_case=False, error_message=None):
        """*Search if one of the strings in a given list exists on the mainframe screen. Raised exception can be overridden with error_message*
           
           *Arguments*
           
           ``list_string``: List of Strings. Example: ${list_of_string}=    string1        string2  
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True 
           
           ``error_message``: Override the default message with a new value.
           
           
           *Examples*
           
               | `Page Should Contain Any String` | ${list_of_string} |
               | `Page Should Contain Any String` | ${list_of_string} | ignore_case=${True} |
               | `Page Should Contain Any String` | ${list_of_string} | error_message=New error message |
        """
        message = 'The strings "' + str(list_string) + '" was not found'
        if error_message: message = error_message
        if ignore_case: list_string = [item.lower() for item in list_string]
        for string in list_string:
            result = self._search_string(string, ignore_case)
            if result: break
        if not result: raise Exception(message)

    @keyword(name=None, tags=("Settings",))
    def page_should_not_contain_any_string(self, list_string, ignore_case=False, error_message=None):
        """*Fails if one or more of the strings in a given list exists on the mainframe screen. Raised exception can be overridden with error_message* 
        
        *Arguments*
           
           ``list_string``: List of Strings. Example: ${list_of_string}=    string1        string2  
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True 
           
           ``error_message``: Override the default message with a new value.
        
        *Examples*
        
            | `Page Should Not Contain Any Strings` | ${list_of_string} |
            | `Page Should Not Contain Any Strings` | ${list_of_string} | ignore_case=${True} |
            | `Page Should Not Contain Any Strings` | ${list_of_string} | error_message=New error message |
        """
        self._compare_all_list_with_screen_text(list_string, ignore_case, error_message, should_match=False)

    @keyword(name=None, tags=("Settings",))
    def page_should_contain_all_strings(self, list_string, ignore_case=False, error_message=None):
        """*Search if all of the strings in a given list exists on the mainframe screen. Raised exception can be overridden with error_message*
        
        *Arguments*
           
           ``list_string``: List of Strings. Example: ${list_of_string}=    string1        string2  
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True 
           
           ``error_message``: Override the default message with a new value.
        
        *Examples*
        
            | `Page Should Contain All Strings` | ${list_of_string} |
            | `Page Should Contain All Strings` | ${list_of_string} | ignore_case=${True} |
            | `Page Should Contain All Strings` | ${list_of_string} | error_message=New error message |
        """
        self._compare_all_list_with_screen_text(list_string, ignore_case, error_message, should_match=True)

    @keyword(name=None, tags=("Settings",))
    def page_should_not_contain_all_strings(self, list_string, ignore_case=False, error_message=None):
        """*Fails if one of the strings in a given list exists on the mainframe screen. Raised exception can be overridden with error_message*
        
        *Arguments*
           
           ``list_string``: List of Strings. Example: ${list_of_string}=    string1        string2  
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True 
           
           ``error_message``: Override the default message with a new value.
        
        *Examples*
        
            | `Page Should Not Contain All Strings` | ${list_of_string} |
            | `Page Should Not Contain All Strings` | ${list_of_string} | ignore_case=${True} |
            | `Page Should Not Contain All Strings` | ${list_of_string} | error_message=New error message |
        """
        message = error_message
        if ignore_case: list_string = [item.lower() for item in list_string]
        for string in list_string:
            result = self._search_string(string, ignore_case)
            if result:
                if message is None:
                    message = 'The string "' + string + '" was found'
                raise Exception(message)

    @keyword(name=None, tags=("Settings",))
    def page_should_contain_string_x_times(self, txt, number, ignore_case=False, error_message=None):
        """*Search if the entered string appears the desired number of times on the mainframe screen. Raised exception can be overridden with error_message*
         
        *Arguments*
           
           ``txt``: String to be matched with 
           
           ``number``: Number of times the string to exist  
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True
           
           ``error_message``: Override the default message with a new value.
        
        *Examples*
        
               | `Page Should Contain String X Times` | something | 3 |
               | `Page Should Contain String X Times` | someTHING | 3 | ignore_case=${True} |
               | `Page Should Contain String X Times` | something | 3 | error_message=New error message |
        """
        message = error_message
        number = int(number)
        all_screen = self._read_all_screen()
        if ignore_case:
            txt = str(txt).lower()
            all_screen = str(all_screen).lower()
        number_of_times = all_screen.count(txt)
        if number_of_times != number:
            if message is None:
                message = 'The string "' + txt + '" was not found "' + str(number) + '" times, it appears "' \
                          + str(number_of_times) + '" times'
            raise Exception(message)
        logger.info('The string "' + txt + '" was found "' + str(number) + '" times')

    @keyword(name=None, tags=("Settings",))
    def page_should_match_regex(self, regex_pattern):
        """*Fails if string does not match pattern as a regular expression.*
        Regular expression check is implemented using the Python [https://docs.python.org/2/library/re.html|re module]. Python's
        regular expression syntax is derived from Perl, and it is thus also very similar to the syntax used,
        for example, in Java, Ruby and .NET.
        Backslash is an escape character in the test data, and possible backslashes in the pattern must
        thus be escaped with another backslash (e.g. \\\d\\\w+).
        """
        page_text = self._read_all_screen()
        if not re.findall(regex_pattern, page_text, re.MULTILINE):
            raise Exception('No matches found for "' + regex_pattern + '" pattern')

    @keyword(name=None, tags=("Settings",))
    def page_should_not_match_regex(self, regex_pattern):
        """*Fails if string does match pattern as a regular expression.* 
        Regular expression check is implemented using the Python [https://docs.python.org/2/library/re.html|re module]. Python's
        regular expression syntax is derived from Perl, and it is thus also very similar to the syntax used,
        for example, in Java, Ruby and .NET.
        Backslash is an escape character in the test data, and possible backslashes in the pattern must
        thus be escaped with another backslash (e.g. \\\d\\\w+).
        """
        page_text = self._read_all_screen()
        if re.findall(regex_pattern, page_text, re.MULTILINE):
            raise Exception('There are matches found for "' + regex_pattern + '" pattern')

    @keyword(name=None, tags=("Settings",))
    def page_should_contain_match(self, txt, ignore_case=False, error_message=None):
        """*Fails unless the given string matches the given pattern. Raised exception can be overridden with error_message*
        Pattern matching is similar as matching files in a shell, and it is always case-sensitive.
        In the pattern, * matches to anything and ? matches to any single character.
        Note that the entire screen is only considered a string for this keyword, so if you want to search
        for the string "something" and it is somewhere other than at the beginning or end of the screen it
        should be reported as follows: **something**
        
        
        *Arguments*
           
           ``txt``: String to be matched with. Can include a pattern. 
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True
           
           ``error_message``: Override the default message with a new value.
        
        *Examples*
        
            | `Page Should Contain Match` | **something** |
            | `Page Should Contain Match` | **so???hing** |
            | `Page Should Contain Match` | **someTHING** | ignore_case=${True} |
            | `Page Should Contain Match` | **something** | error_message=New error message |
        """
        message = error_message
        all_screen = self._read_all_screen()
        if ignore_case:
            txt = txt.lower()
            all_screen = all_screen.lower()
        matcher = Matcher(txt, caseless=False, spaceless=False)
        result = matcher.match(all_screen)
        if not result:
            if message is None:
                message = 'No matches found for "' + txt + '" pattern'
            raise Exception(message)

    @keyword(name=None, tags=("Settings",))
    def page_should_not_contain_match(self, txt, ignore_case=False, error_message=None):
        """*Fails if the given string matches the given pattern. Raised exception can be overridden with error_message*
        Pattern matching is similar as matching files in a shell, and it is always case-sensitive.
        In the pattern, * matches to anything and ? matches to any single character.
        Note that the entire screen is only considered a string for this keyword, so if you want to search
        for the string "something" and it is somewhere other than at the beginning or end of the screen it
        should be reported as follows: **something**
        The search is case sensitive, if you want ignore this you can pass the argument: ignore_case=${True} and you
        can edit the raise exception message with error_message.
        
        *Arguments*
           
           ``txt``: String to be matched with. Can include a pattern. 
           
           ``ignore_case``: Default is Case-sensitive. To ignore case, pass the argument: ignore_case=True
           
           ``error_message``: Override the default message with a new value.
        
        *Examples*
        
            | `Page Should Not Contain Match` | **something** |
            | `Page Should Not Contain Match` | **so???hing** |
            | `Page Should Not Contain Match` | **someTHING** | ignore_case=${True} |
            | `Page Should Not Contain Match` | **something** | error_message=New error message |
        """
        message = error_message
        all_screen = self._read_all_screen()
        if ignore_case:
            txt = txt.lower()
            all_screen = all_screen.lower()
        matcher = Matcher(txt, caseless=False, spaceless=False)
        result = matcher.match(all_screen)
        if result:
            if message is None:
                message = 'There are matches found for "' + txt + '" pattern'
            raise Exception(message)

    @keyword(name=None, tags=("Settings",))
    def _read_all_screen(self):
        """*Read all the mainframe screen and return in a single string.*
        """
        full_text = ''
        for ypos in range(24):
            line = self.mf.string_get(ypos + 1, 1, 80)
            for char in line:
                if char:
                    full_text += char
        return full_text
		
    @keyword(name=None,  tags=("Settings",))
    def login(self, userID, password):
        """*Authenticate to Mainframe*
        
        *Arguments*
           
           ``userID``: AB ID of user with mainframe access
           
           ``password``: Credentail associated to the AB ID of the user
                   
        *Examples*
        
            | `Login` | AB00000 | test1234 |
        """
        logger.debug('Logging In with UserID:' + userID + ' and Password:' + password)
        self._write(userID)
        self.mf.exec_command(b'Tab')	
        self._write(password)
        self._write("Yes", "019", "020")				
        self.mf.send_enter()
        time.sleep(self.wait)
        status = self._search_string("Session Description")
        if status:
            logger.info('Logged In successfully')
        if not status:
            raise Exception('Unable to authenticate for unknown reason, try logging in manually with same credentails')
	
    @keyword(name=None,  tags=("Settings",))
    def choose_session(self, session_id):
        """*Choose the required session by providing sessid*
        
        *Arguments*
           
           ``session_id``: Name of the session in CAPS
                   
        *Examples*
        
            | `Choose Session` | PUCICA8 |
        """
        message = 'Session "' + session_id + '" was not found'
        result, xpos, ypos = self._search_string_and_return_coordinates("_ "+session_id)
        if result:
            logger.debug('Found Session ' + session_id)
            self._write("i", ypos=ypos+1, xpos=xpos+1)
            self.mf.send_enter()
            time.sleep(self.wait)
            status = self._search_string('has ended')
            if status:
                logger.debug('Ended existing session, if any')						
            self._write("s", ypos=ypos+1, xpos=xpos+1)
            self.mf.send_enter()
            time.sleep(self.wait)
            status = self._search_string("Session Description")
            if not status:
                logger.debug('Session ' +session_id + ' selected') 
            
			# Additional check to handle "atso"
            acf_status = self._search_string("ENTER PROC NAME")
            if acf_status:
                logger.debug('Landed on ACF Screen; Entered "atso"')
                self._write("atso")
                self.mf.send_enter()
                self.mf.send_enter()
                self.mf.send_enter()
                time.sleep(self.wait)
				
            login_status = self._search_string("LOGON IN PROGRESS")
            if login_status:
                logger.debug('Landed on Login In Progress Screen')
                self.mf.send_enter()
                time.sleep(self.wait)
                ispf_status = self._search_string("ISPF Primary Option Menu")
                if ispf_status:
                    logger.debug('Landed on ISPF screen')

            logger.info("Session " + session_id + " selected")
			
        if not result: 
            #TODO: Search Next pages until last page and raise exception if session is not found
            raise Exception(message)
        
	
    
    def _search_string_and_return_coordinates(self, string):
        """*Search if the string exists on the mainframe screen and return if True or False along with x and y coordinates.*
           
        """
        def __read_screen_and_return_coordinates(string):
            xpos = 0
            for ypos in range(24):
                line = self.mf.string_get(ypos+1, 1, 80)
                if string in line:
                    logger.debug('The string "' + string + '" was found, returned True, x and y coordinates')
                    xpos = line.index(string)
                    return True, xpos, ypos
            return False, xpos, ypos
        status, xpos, ypos = __read_screen_and_return_coordinates(string)
        return status, xpos, ypos
	
		
    @keyword(name=None,  tags=("Settings",))
    def set_ispf_options(self, ispf_options):
        """*Choose ISPF Primary Menu Options*
        
        *Arguments*
           
           ``ispf_options``: Menu option. For ex: 3.4
                   
        *Examples*
        
            | `Set Ispf Options` | 3.4 |
        """
        ispf_status = self._search_string("ISPF Primary Option Menu")
        if ispf_status:
            self._write(ispf_options)
            self.mf.send_enter()
            time.sleep(self.wait)
            logger.info('ISPF Menu Option ' + ispf_options + ' selected')
            return True
        if not ispf_status:
            logger.info('Cursor not on ISPF screen, cannot set options')
            return False
		
    @keyword(name=None,  tags=("Settings",))
    def search_pds(self, full_qualifier_name):
        """*Search PDS by providing full qualifier name*
        
        *Arguments*
           
           ``full_qualifier_name``: Full qualifier name. For ex: UTILITY.TESTJCL
                   
        *Examples*
        
            | `Search Pds` | UTILITY.TESTJCL |
        """
        # Clear Dsname Level field in Data Set List Utility screen
        self.delete_field("009", "024")
        time.sleep(self.wait)	
        self._write(full_qualifier_name, "009", "024")
        self.mf.send_enter()
        time.sleep(self.wait)
		
		#Exception Handling - throw exception if dataset not found 
        result = self._search_string("No data set names found")
        if result:
            raise Exception('No data set names found for qualifier: "' + full_qualifier_name + '"')		
        
        if not result:
            logger.info("Retrieved dataset: " +full_qualifier_name)

    @keyword(name=None,  tags=("Settings",))
    def view_pds_members(self):
        """*View members of data set*
                          
        *Examples*
        
            | `View Pds Members` | 
        """
        self._write("m", "007", "002")				
        self.mf.send_enter()
        time.sleep(self.wait)

        #Exception Handling 
        result = self._search_string("AUTHORIZATION IS REQUIRED")
        if result:
            raise Exception('DATASET CANNOT BE OPENED; AUTHORIZATION IS REQUIRED')
        if not result:
            logger.info('No authorization issue; assuming pds members are viewed')
            		
			
    @keyword(name=None,  tags=("Settings",))
    def submit_jcl(self, jcl_name=None):
        """*Submit a JCL and returns the JOB ID.*
                          
        *Arguments*
           
           ``jcl_name``: Optional. If not provided, the first job in the list is executed
                   		
		*Examples*
        
            | `Submit Jcl` | 
            | `Submit Jcl` | P2380SFN | 
			| `Submit Jcl` | P2380SFN | 4 | # PF(3) is executed 'x' number of times based on the count provided
        """
        if jcl_name:
            result, xpos, ypos = self._search_string_and_return_coordinates("_ "+jcl_name)
            if result:
                self._write("j", ypos=ypos+1, xpos=xpos+1)
                self.mf.send_enter()
                time.sleep(self.wait)
            if not result:
                raise Exception('JCL "' + jcl_name + '", not found, cannot submit JOB "')					
        
        if jcl_name is None:
            self._write("j", "005", "003")				
            self.mf.send_enter()
            time.sleep(self.wait)
        
        result = self._search_string("SUBMITTED")
        if result:
            logger.info("JOB submitted successfully")
            page_text = self._read_all_screen()
            jobDetails = page_text.partition("JOB")[2]
            jobIDText = jobDetails.partition("JOB")[2]
            jobID = "JOB" + jobIDText[:5]
            logger.info("JOB Details: " +jobDetails) 
            self.mf.send_enter()
            time.sleep(self.wait)
            return jobID
        if not result:
            raise Exception('Something went wrong, JOB cannot be submitted')		
			
    @keyword(name=None,  tags=("Settings",))
    def assert_job_status(self, jobID, max_rc):
        """*Asserts that the return code is expected*
		
		*Arguments*
           
           ``jobID``: JOB ID captured in `Submit Jcl`  
		   
		   ``max_rc``: Expected return code 
           
           ``go_back_count``: Optional. If not provided, cursor lands on ISPF Options Menu Screen
                          
        *Examples*
        
            | `Assert Job Status` | JOB07306 | CC 0000 |
			| `Assert Job Status` | JOB07306 | CC 0000 | 2 |
        """
        ispf_status = self._search_string("ISPF Primary Option Menu")
        if not ispf_status:
            raise Exception('Cannot Assert, cursor not on ISPF menu screen')
        if ispf_status:
            self._write("s;st")				
            self.mf.send_enter()
            time.sleep(self.wait)
            self.mf.send_enter()
            time.sleep(self.wait)			
            totalLines = self.read("003", "068", "2")
            currentLine = self.read("003", "064", "2")
            
            while (currentLine <= totalLines):
        
                status = self._find_jobID_on_screen(jobID)
        
                if status:
                    #TODO: Extract Max-RC Status of JOB
                    result, xpos, ypos = self._search_string_and_return_coordinates(jobID)
                    maxRCReturned = self.read(ypos+1, xpos+9, "11")
                    
                    if max_rc == maxRCReturned.strip():
                        return True
                        
                    if max_rc != maxRCReturned.strip():
                        logger.info('Expected Max-RC "' + max_rc + '", instead found "' + maxRCReturned.strip() + '" ')
                        raise Exception('Assertion Failed, expected and actual max-rc not matching')
                    break
                if not status:
                    self.execute_command("PF(8)")
                    time.sleep(self.wait)
                    newLine = self.read("003", "064", "2")
                    logger.info(newLine)
                    logger.info(currentLine)
                    if newLine == currentLine:
                       raise Exception('Job ID not found in spool')
                    if newLine != currentLine:
                        currentLine = newLine
                    continue            
        
    def _find_jobID_on_screen(self, jobID):
        page_text = self._read_all_screen()
        if jobID in page_text:
            logger.debug('Found JOB ID on screen')
            return True
        if jobID not in page_text:
            return False           

    @keyword(name=None,  tags=("Settings",))
    def return_to_screen(self, screen_identifier, returnKey):
        """*Returns the cursor back to desired screen*
		
		*Arguments*
           
           ``screen_identifier``: Unique Text Identifier on the expected screen. For Ex: ISPF
		   
		   ``returnKey``: Function Key that returns the cursor to previous screen. For Ex: PF(3) 
                          
        *Examples*
        
            | `Return To Screen` | ISPF | PF(3) |
        """
        status = 1
        while (status != 0):
            search_status = self._search_string(screen_identifier)
            if search_status:
                status = 0
                break
            if not search_status:
                self.execute_command(returnKey)
                time.sleep(self.wait)
                continue
				
            
    def _compare_all_list_with_screen_text(self, list_string, ignore_case, message, should_match):
        if ignore_case: list_string = [item.lower() for item in list_string]
        for string in list_string:
            result = self._search_string(string, ignore_case)
            if not should_match and result:
                if message is None:
                    message = 'The string "' + string + '" was found'
                raise Exception(message)
            elif should_match and not result:
                if message is None:
                    message = 'The string "' + string + '" was not found'
                raise Exception(message)

    @staticmethod
    def _check_limits(ypos, xpos):
        """Checks if the user has passed some coordinate y / x greater than that existing in the mainframe
        """
        if int(ypos) > 24:
            raise Exception('You have exceeded the y-axis limit of the mainframe screen')
        if int(xpos) > 80:
            raise Exception('You have exceeded the x-axis limit of the mainframe screen')




from .buttons import ButtonKeywords  # noqa: F401 #pylint: disable=unused-import
from .label import LabelKeywords  # noqa: F401 #pylint: disable=unused-import
from .listcontrols import ListKeywords  # noqa: F401 #pylint: disable=unused-import
from .listview import ListViewKeywords  # noqa: F401 #pylint: disable=unused-import
from .menu import MenuKeywords  # noqa: F401 #pylint: disable=unused-import
from .mouse import MouseKeywords  # noqa: F401 #pylint: disable=unused-import
from .progressbar import ProgressbarKeywords  # noqa: F401 #pylint: disable=unused-import
from .slider import SliderKeywords  # noqa: F401 #pylint: disable=unused-import
from .tab import TabKeywords  # noqa: F401 #pylint: disable=unused-import
from .textbox import TextBoxKeywords  # noqa: F401 #pylint: disable=unused-import
from .tree import TreeKeywords  # noqa: F401 #pylint: disable=unused-import
from .uiitem import UiItemKeywords  # noqa: F401 #pylint: disable=unused-import



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import ftplib
import os
import socket
from robot.api import logger
from robot.api.deco import keyword

class rbfFTP(object):

    """
This library provides functionality of FTP client.


The simplest example (connect, change working dir, print working dir, close):
 | ftp connect | 192.168.1.10 | mylogin | mypassword |
 | cwd | /home/myname/tmp/testdir |
 | pwd |
 | ftp close |

It is possible to use multiple ftp connections in parallel. Connections are
identified by string identifiers:
 | ftp connect | 192.168.1.10 | mylogin | mypassword | connId=ftp1 |
 | ftp connect | 192.168.1.20 | mylogin2 | mypassword2 | connId=ftp2 |
 | cwd | /home/myname/tmp/testdir | ftp1 |
 | cwd | /home/myname/tmp/testdir | ftp2 |
 | pwd | ftp2 |
 | pwd | ftp1 |
 | ftp close | ftp2 |
 | ftp close | ftp1 |

To run library remotely execute: python rbfFTP.py <ipaddress> <portnumber>
(for example: python rbfFTP.py 192.168.0.101 8222)
"""

    def __getConnection(self, connId):
        if connId in self.ftpList:
            return self.ftpList[connId]
        else:
            errMsg = "Connection with ID %s does not exist. It should be created before this step." % connId
            raise rbfFTPError(errMsg)

    def __addNewConnection(self, connObj, connId):
        if connId in self.ftpList:
            errMsg = "Connection with ID %s already exist. It should be deleted before this step." % connId
            raise rbfFTPError(errMsg)
        else:
            self.ftpList[connId] = connObj

    def __removeConnection(self, connId):
        if connId in self.ftpList:
            self.ftpList.pop(connId)

    @keyword
    def getAllFtpConnections(self):
        """
        *Returns a dictionary containing active ftp connections.*
        """
        outputMsg = "Current ftp connections:\n"
        counter = 1
        for k in self.ftpList:
            outputMsg += str(counter) + ". " + k + " "
            outputMsg += str(self.ftpList[k]) + "\n"
            counter += 1
        if self.printOutput:
            logger.info(outputMsg)
        return self.ftpList

    @keyword
    def ftp_connect(self, host, user='anonymous', password='anonymous@', port=21, timeout=30, connId='default'):
        """
        *Constructs FTP object, opens a connection and login. Returns server output.*
        Call this function before any other (otherwise raises exception).
        
        
        *Arguments*
        
            ``host``: server host address
            
            ``user``: Optional. FTP user name. If not given, 'anonymous' is used.
            
            ``password``: Optional. FTP password. If not given, 'anonymous@' is used.
            
            ``port``: Optional. TCP port. By default 21.
            
            ``timeout``: Optional. Timeout in seconds. By default 30.
            
            ``connId``: Optional. Connection identifier. By default equals 'default'
            
        *Examples*
        
        | `ftp connect` | 192.168.1.10 | mylogin | mypassword |  |  |
        | `ftp connect` | 192.168.1.10 |  |  |  |  |
        | `ftp connect` | 192.168.1.10 | mylogin | mypassword | connId=secondConn |  |
        | `ftp connect` | 192.168.1.10 | mylogin | mypassword | 29 | 20 |
        | `ftp connect` | 192.168.1.10 | mylogin | mypassword | 29 |  |
        | `ftp connect` | 192.168.1.10 | mylogin | mypassword | timeout=20 |  |
        | `ftp connect` | 192.168.1.10 | port=29 | timeout=20 |  |  |
        """
        if connId in self.ftpList:
            errMsg = "Connection with ID %s already exist. It should be deleted before this step." % connId
            raise rbfFTPError(errMsg)
        else:
            newFtp = None
            outputMsg = ""
            try:
                timeout = int(timeout)
                port = int(port)
                newFtp = ftplib.FTP()
                outputMsg += newFtp.connect(host, port, timeout)
                outputMsg += newFtp.login(user,password)
            except socket.error as se:
                raise rbfFTPError('Socket error exception occured.')
            except ftplib.all_errors as e:
                raise rbfFTPError(str(e))
            except Exception as e:
                raise rbfFTPError(str(e))
            if self.printOutput:
                logger.info(outputMsg)
            self.__addNewConnection(newFtp, connId)

    @keyword
    def get_welcome(self, connId='default'):
        """
        *Returns wlecome message of FTP server.*
        
        *Arguments*
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
            
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            outputMsg += thisConn.getwelcome()
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def pwd(self, connId='default'):
        """
        *Returns the pathname of the current directory on the server.*
        
        *Arguments*
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            outputMsg += thisConn.pwd()
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def cwd(self, directory, connId='default'):
        """
        *Changes working directory and returns server output.* 
        
        *Arguments*
        
        ``directory``: A path to which working dir should be changed.
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        
        *Examples*
        
        | `cwd` | /home/myname/tmp/testdir |  |
        | `cwd` | /home/myname/tmp/testdir | ftp1 |
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            outputMsg += thisConn.cwd(directory)
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    @keyword
    def dir(self, connId='default'):
        """
        *Returns list of raw lines returned as contens of current directory.*
        
        *Arguments*
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        """
        dirList = []
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            thisConn.dir(dirList.append)
            for d in dirList:
                outputMsg += str(d) + "\n"
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return dirList

    
    @keyword
    def dir_names(self, connId='default'):
        """
        *Returns list of files (and/or directories) of current directory.*
        
        *Arguments*
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        """
        files_list = []
        thisConn = self.__getConnection(connId)
        try:
            files_list = thisConn.nlst()
        except:
            files_list = []
        return files_list

    
    @keyword
    def mkd(self, newDirName, connId='default'):
        """
        *Creates new directory on FTP server. Returns new directory path.*
        
        *Arguments*
        
        ``newDirName``: Name of new directory
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            outputMsg += str(thisConn.mkd(newDirName))
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def rmd(self, directory, connId='default'):
        """
        *Deletes directory from FTP server. Returns server output.*
        
        *Arguments*
        
        ``directory``: Path to a directory to be deleted
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            outputMsg += str(thisConn.rmd(directory))
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def download_file(self, remoteFileName, localFilePath=None, connId='default'):
        """
        *Downloads file from current directory on FTP server in binary mode. Returns server output.*
        If localFilePath is not given, file is saved in current local directory (by
        default folder containing robot framework project file) with the same name
        as source file. 
        
        *Arguments*
        
        ``remoteFileName``: File name on FTP server
        
        ``localFilePath``: Optional. Local file name or path where remote file should be saved. localFilePath variable can have following meanings:
        - file name (will be saved in current default directory);
        - full path (dir + file name)
        - dir path (original file name will be added)
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
                
        *Examples*
        
        | `download file` | a.txt |  |  |
        | `download file` | a.txt | b.txt | connId=ftp1 |
        | `download file` | a.txt | D:/rfftppy/tmp |  |
        | `download file` | a.txt | D:/rfftppy/tmp/b.txt |  |
        | `download file` | a.txt | D:\\rfftppy\\tmp\\c.txt |  |
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        localPath = ""
        if localFilePath == None:
            localPath = remoteFileName
        else:
            localPath = os.path.normpath(localFilePath)
            if os.path.isdir(localPath):
                localPath = os.path.join(localPath, remoteFileName)
        try:
            with open(localPath, 'wb') as localFile:
                outputMsg += thisConn.retrbinary("RETR " + remoteFileName, localFile.write)
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def upload_file(self, localFileName, remoteFileName=None, connId='default'):
        """
        *Sends file from local drive to current directory on FTP server in binary mode. Returns server output.*
        
        
        *Arguments*
        
        ``localFileName``: File name or path to a file on a local drive.
        
        ``remoteFileName``: Optional. A name or path containing name under which file should be saved.
        If remoteFileName agument is not given, local name will be used.
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
       
        *Examples*
        
        | `upload file` | x.txt | connId=ftp1 |
        | `upload file` | D:/rfftppy/y.txt |  |
        | `upload file` | u.txt | uu.txt |
        | `upload file` | D:/rfftppy/z.txt | zz.txt |
        | `upload file` | D:\\rfftppy\\v.txt |  |
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        remoteFileName_ = ""
        localFilePath = os.path.normpath(localFileName)
        if not os.path.isfile(localFilePath):
            raise rbfFTPError("Valid file path should be provided.")
        else:
            if remoteFileName==None:
                fileTuple = os.path.split(localFileName)
                if len(fileTuple)==2:
                    remoteFileName_ = fileTuple[1]
                else:
                    remoteFileName_ = 'defaultFileName'
            else:
                remoteFileName_ = remoteFileName
            try:
                outputMsg += thisConn.storbinary("STOR " + remoteFileName_, open(localFilePath, "rb"))
            except ftplib.all_errors as e:
               raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def size(self, fileToCheck, connId='default'):
        """
        *Checks size of a file on FTP server. Returns size of a file in bytes (integer).*
        Note that the SIZE command is not standardized, but is supported by many common server implementations.
        
        *Arguments*
        
        ``fileToCheck``: File name or path to a file on FTP server
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        
        *Examples*
        
        | ${file1size} = | `size` | /home/myname/tmp/uu.txt | connId=ftp1 |
        | Should Be Equal As Numbers | ${file1size} | 31 |  |

        
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            tmpSize = thisConn.size(fileToCheck)
            outputMsg += str(tmpSize)
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def rename(self, targetFile, newName, connId='default'):
        """
        *Renames (actually moves) file on FTP server. Returns server output.*
        
        
        *Arguments*
        
        ``targetFile``: Name of a file or path to a file to be renamed
        
        ``newName``: New name or new path
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        
        *Arguments*
        
        | `rename` | tmp/z.txt | /home/myname/tmp/testdir/z.txt |
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            outputMsg += str(thisConn.rename(targetFile, newName))
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def delete(self, targetFile, connId='default'):
        """
        *Deletes file on FTP server. Returns server output.*
        
        *Arguments*
        
        ``targetFile``: Name of a file or path to a file to be renamed
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            outputMsg += str(thisConn.delete(targetFile))
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def send_cmd(self, command, connId='default'):
        """
        *Sends any command to FTP server. Returns server output.*
        
        *Arguments*
        
        ``command``: Any valid command to be sent (invalid will result in exception).
        
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        *Examples*
        
        | `send cmd` | HELP |
        """
        thisConn = self.__getConnection(connId)
        outputMsg = ""
        try:
            outputMsg += str(thisConn.sendcmd(command))
        except ftplib.all_errors as e:
            raise rbfFTPError(str(e))
        if self.printOutput:
            logger.info(outputMsg)
        return outputMsg

    
    @keyword
    def ftp_close(self, connId='default'):
        """
        *Closes FTP connection. Returns None.*
        
        *Arguments*
              
        ``connId``: Optional. Connection identifier. By default equals 'default'
        
        """
        thisConn = self.__getConnection(connId)
        try:
            thisConn.quit()
            self.__removeConnection(connId)
        except Exception as e:
            try:
                thisConn.close()
                self.__removeConnection(connId)
            except ftplib.all_errors as x:
                raise rbfFTPError(str(x))

    def __del__(self):
        self.ftpList = {}

class rbfFTPError(Exception):
    def __init__(self,msg):
        self.msg = msg

    def __str__(self):
        return self.msg

def main():
    import sys
    from robotremoteserver import RobotRemoteServer
    print("Starting Robot Framework Ftp Library as a remote server ...")
    RobotRemoteServer(library=rbfFTP(), host=sys.argv[1], port=sys.argv[2])

if __name__ == '__main__':
    main()



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from robot.libraries.BuiltIn import BuiltIn

from rbfWEB.base import ContextAware


try:
    basestring
except NameError:
    basestring = str


class CustomLocator(ContextAware):

    def __init__(self, ctx, name, finder):
        ContextAware.__init__(self, ctx)
        self.name = name
        self.finder = finder

    def find(self, criteria, tag, constraints, parent):
        # Allow custom locators to be keywords or normal methods
        if isinstance(self.finder, basestring):
            element = BuiltIn().run_keyword(self.finder, parent,
                                            criteria, tag, constraints)
        elif hasattr(self.finder, '__call__'):
            element = self.finder(parent, criteria, tag, constraints)
        else:
            raise AttributeError('Invalid type provided for Custom Locator %s'
                                 % self.name)

        # Always return an array
        if hasattr(element, '__len__') and not isinstance(element, basestring):
            return element
        else:
            return [element]



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from robot.api import logger
from robot.api.deco import keyword



class Query(object):
    """
    Query handles all the querying done by the Database Library.
    """

    
    @keyword(name=None, tags=("Query",))
    def query(self, selectStatement, sansTran=False, returnAsDict=False):
        """
        *Uses the input `selectStatement` to query for the values that will be returned as a list of tuples.*
        
        Tip: Unless you want to log all column values of the specified rows, try specifying the column names in your select statements as much as possible to prevent any unnecessary surprises with schema
        changes and to easily see what your [] indexing is trying to retrieve
        (i.e. instead of `"select * from my_table"`, try `"select id, col_1, col_2 from my_table"`).
        
        *Arguments*

        ``selectStatement``: SELECT query 

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        ``returnAsDict``: Optional. Set `returnAsDict` to True to return values as a list of dictionaries.
        
        *Examples*

        For example, given we have a table `person` with the following data:
        | id | first_name  | last_name |
        |  1 | Franz Allan | See       |

        When you do the following:
        | @{queryResults} | `Query` | SELECT * FROM person |
        | Log Many | @{queryResults} |

        You will get the following:
        [1, 'Franz Allan', 'See']

        Also, you can do something like this:
        | ${queryResults} | `Query` | SELECT first_name, last_name FROM person |
        | Log | ${queryResults[0][1]}, ${queryResults[0][0]} |

        And get the following
        See, Franz Allan

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | @{queryResults} | `Query` | SELECT * FROM person | True |
        """
        cur = None
        try:
            cur = self._dbconnection.cursor()
            logger.info('Executing : Query  |  %s ' % selectStatement)
            self.__execute_sql(cur, selectStatement)
            allRows = cur.fetchall()

            if returnAsDict:
                mappedRows = []
                col_names = [c[0] for c in cur.description]

                for rowIdx in range(len(allRows)):
                    d = {}
                    for colIdx in range(len(allRows[rowIdx])):
                        d[col_names[colIdx]] = allRows[rowIdx][colIdx]
                    mappedRows.append(d)
                return mappedRows

            return allRows
        finally:
            if cur:
                if not sansTran:
                    self._dbconnection.rollback()

    
    
    @keyword(name=None, tags=("Query",))
    def row_count(self, selectStatement, sansTran=False):
        """
        *Uses the input `selectStatement` to query the database and returns the number of rows from the query.*
        
        *Arguments*

        ``selectStatement``: SELECT query 

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.

        *Examples*
        
        For example, given we have a table `person` with the following data:
        | id | first_name  | last_name |
        |  1 | Franz Allan | See       |
        |  2 | Jerry       | Schneider |

        When you do the following:
        | ${rowCount} | `Row Count` | SELECT * FROM person |
        | Log | ${rowCount} |

        You will get the following:
        2

        Also, you can do something like this:
        | ${rowCount} | `Row Count` | SELECT * FROM person WHERE id = 2 |
        | Log | ${rowCount} |

        And get the following
        1

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | ${rowCount} | `Row Count` | SELECT * FROM person | True |
        """
        cur = None
        try:
            cur = self._dbconnection.cursor()
            logger.info('Executing : Row Count  |  %s ' % selectStatement)
            self.__execute_sql(cur, selectStatement)
            data = cur.fetchall()
            if self.db_api_module_name in ["sqlite3", "ibm_db", "ibm_db_dbi", "pyodbc"]:
                rowCount = len(data)
            else:
                rowCount = cur.rowcount
            return rowCount
        finally:
            if cur:
                if not sansTran:
                    self._dbconnection.rollback()

    
    @keyword(name=None, tags=("Query",))
    def description(self, selectStatement, sansTran=False):
        """
        *Uses the input `selectStatement` to query a table in the db which will be used to determine the description.*
        
        *Arguments*

        ``selectStatement``: SELECT query 

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.

        *Examples*

        For example, given we have a table `person` with the following data:
        | id | first_name  | last_name |
        |  1 | Franz Allan | See       |

        When you do the following:
        | @{queryResults} | `Description` | SELECT * FROM person |
        | Log Many | @{queryResults} |

        You will get the following:
        [Column(name='id', type_code=1043, display_size=None, internal_size=255, precision=None, scale=None, null_ok=None)]
        [Column(name='first_name', type_code=1043, display_size=None, internal_size=255, precision=None, scale=None, null_ok=None)]
        [Column(name='last_name', type_code=1043, display_size=None, internal_size=255, precision=None, scale=None, null_ok=None)]

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | @{queryResults} | `Description` | SELECT * FROM person | True |
        """
        cur = None
        try:
            cur = self._dbconnection.cursor()
            logger.info('Executing : Description  |  %s ' % selectStatement)
            self.__execute_sql(cur, selectStatement)
            description = list(cur.description)
            if sys.version_info[0] < 3:
                for row in range(0, len(description)):
                    description[row] = (description[row][0].encode('utf-8'),) + description[row][1:]
            return description
        finally:
            if cur:
                if not sansTran:
                    self._dbconnection.rollback()

    
    @keyword(name=None, tags=("Query",))
    def delete_all_rows_from_table(self, tableName, sansTran=False):
        """
        *Delete all the rows within a given table.*
        
        *Arguments*

        ``tableName``: Name of the table

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples*
        
        For example, given we have a table `person` in a database

        When you do the following:
        | `Delete All Rows From Table` | person |

        If all the rows can be successfully deleted, then you will get:
        | `Delete All Rows From Table` | person | # PASS |
        
        If the table doesn't exist or all the data can't be deleted, then you will get:
        | `Delete All Rows From Table` | first_name | # FAIL |

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Delete All Rows From Table` | person | True |
        """
        cur = None
        selectStatement = ("DELETE FROM %s;" % tableName)
        try:
            cur = self._dbconnection.cursor()
            logger.info('Executing : Delete All Rows From Table  |  %s ' % selectStatement)
            result = self.__execute_sql(cur, selectStatement)
            if result is not None:
                if not sansTran:
                    self._dbconnection.commit()
                return result
            if not sansTran:
                self._dbconnection.commit()
        finally:
            if cur:
                if not sansTran:
                    self._dbconnection.rollback()

    
    @keyword(name=None, tags=("Query",))
    def execute_sql_script(self, sqlScriptFileName, sansTran=False):
        """
        *Executes the content of the `sqlScriptFileName` as SQL commands and returns number of rows affected. Useful for setting the database to a known state before running your tests, or clearing out your test
        data after running each a test.* 
        
        *Arguments*

        ``sqlScriptFileName``: Name and location of the script file to be executed.

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples*

        | `Execute Sql Script` | ${EXECDIR}${/}resources${/}DDL-setup.sql |
        | `Execute Sql Script` | ${EXECDIR}${/}resources${/}DML-setup.sql |
        | `Execute Sql Script` | ${EXECDIR}${/}resources${/}DML-teardown.sql |
        | `Execute Sql Script` | ${EXECDIR}${/}resources${/}DDL-teardown.sql |

        SQL commands are expected to be delimited by a semi-colon (';').

        For example:
        DELETE FROM person_employee_table;
        DELETE FROM person_table;
        DELETE FROM employee_table;

        Also, the last SQL command can optionally omit its trailing semi-colon.

        For example:
        DELETE FROM person_employee_table;
        DELETE FROM person_table;
        DELETE FROM employee_table

        Given this, that means you can create spread your SQL commands in several lines.

        For example:
        DELETE
          FROM person_employee_table;
        DELETE
          FROM person_table;
        DELETE
          FROM employee_table

        However, lines that starts with a number sign (`#`) are treated as a
        commented line. Thus, none of the contents of that line will be executed.

        For example:
        # Delete the bridging table first...
        DELETE
          FROM person_employee_table;
          # ...and then the bridged tables.
        DELETE
          FROM person_table;
        DELETE
          FROM employee_table

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Execute Sql Script` | ${EXECDIR}${/}resources${/}DDL-setup.sql | True |
        """
        sqlScriptFile = open(sqlScriptFileName)

        cur = None
        result = 0
        try:
            cur = self._dbconnection.cursor()
            logger.info('Executing : Execute SQL Script  |  %s ' % sqlScriptFileName)
            sqlStatement = ''
            for line in sqlScriptFile:
                PY3K = sys.version_info >= (3, 0)
                if not PY3K:
                    #spName = spName.encode('ascii', 'ignore')
                    line = line.strip().decode("utf-8")
                if line.startswith('#'):
                    continue
                elif line.startswith('--'):
                    continue

                sqlFragments = line.split(';')
                if len(sqlFragments) == 1:
                    sqlStatement += line + ' '
                else:
                    for sqlFragment in sqlFragments:
                        sqlFragment = sqlFragment.strip()
                        if len(sqlFragment) == 0:
                            continue

                        sqlStatement += sqlFragment + ' '

                        result = result + self.__execute_sql(cur, sqlStatement)
                        sqlStatement = ''

            sqlStatement = sqlStatement.strip()
            if len(sqlStatement) != 0:
                result = self.__execute_sql(cur, sqlStatement)

            if not sansTran:
                self._dbconnection.commit()
        finally:
            if cur:
                if not sansTran:
                    self._dbconnection.rollback()
        return result

    
    @keyword(name=None, tags=("Query",))
    def execute_sql_string(self, sqlString, sansTran=False):
        """
        *Executes the sqlString as SQL commands  and returns number of rows affected. Useful to pass arguments to your sql.* 
        
        *Arguments*

        ``sqlString``: SQL statement to be executed.

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples*        

        SQL commands are expected to be delimited by a semi-colon (';').

        For example:
        | `Execute Sql String` | DELETE FROM person_employee_table; DELETE FROM person_table |

        For example with an argument:
        | `Execute Sql String` | SELECT * FROM person WHERE first_name = ${FIRSTNAME} |

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | `Execute Sql String` | DELETE FROM person_employee_table; DELETE FROM person_table | True |
        """
        cur = None
        result = 0
        try:
            cur = self._dbconnection.cursor()
            logger.info('Executing : Execute SQL String  |  %s ' % sqlString)
            result = self.__execute_sql(cur, sqlString)
            if not sansTran:
                self._dbconnection.commit()
        finally:
            if cur:
                if not sansTran:
                    self._dbconnection.rollback()
        return result

    
    @keyword(name=None, tags=("Query",))
    def call_stored_procedure(self, spName, spParams=None, sansTran=False):
        """
        *Uses the inputs of `spName` and 'spParams' to call a stored procedure.* The return from this keyword will always be a list.
        
        *Arguments*

        ``spName``: Name of stored procedure
        
        ``spParams``: List of parameters being sent in. The list can be one or multiple items.

        ``sansTran``: Optional. Set `sansTran` to True to run command without an explicit transaction commit or rollback.
        
        *Examples*

        | @{ParamList} = | Create List | FirstParam | SecondParam | ThirdParam |
        | @{QueryResults} = | `Call Stored Procedure` | DBName.SchemaName.StoredProcName | List of Parameters |

        | @{ParamList} = | Create List | Testing | LastName |
        | Set Test Variable | ${SPName} = | DBTest.DBSchema.MyStoredProc |
        | @{QueryResults} = | `Call Stored Procedure` | ${SPName} | ${ParamList} |
        | Log List | @{QueryResults} |

        Using optional `sansTran` to run command without an explicit transaction commit or rollback:
        | @{QueryResults} = | `Call Stored Procedure` | DBName.SchemaName.StoredProcName | List of Parameters | True |
        """
        if spParams is None:
            spParams = []
        cur = None
        try:
            if self.db_api_module_name in ["cx_Oracle"]:
                cur = self._dbconnection.cursor()
            else:
                cur = self._dbconnection.cursor(as_dict=False)
            PY3K = sys.version_info >= (3, 0)
            if not PY3K:
                spName = spName.encode('ascii', 'ignore')
            logger.info('Executing : Call Stored Procedure  |  %s  |  %s ' % (spName, spParams))
            cur.callproc(spName, spParams)
            cur.nextset()
            retVal=list()
            for row in cur:
                #logger.info ( ' %s ' % (row))
                retVal.append(row)
            if not sansTran:
                self._dbconnection.commit()
            return retVal
        finally:
            if cur:
                if not sansTran:
                    self._dbconnection.rollback()

    def __execute_sql(self, cur, sqlStatement):
        return cur.execute(sqlStatement)



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Originally based on  Robot Framework 3.0.2 robot.utils.robottypes and
# Can be removed when library minimum required Robot Framework
# version is greater than 3.0.4. Then Robot Framework is_truthy should
# also support string NONE as Python False. Also string 0 is evaluated
# as False in RF 3.1 but not in 3.0.4.
from robot.utils import is_string
from robot.utils import PY3


def is_truthy(item):
    if is_string(item):
        return item.upper() not in ('FALSE', 'NO', '', 'NONE', '0', 'OFF')
    return bool(item)


def is_falsy(item):
    return not is_truthy(item)


def is_noney(item):
    return item is None or is_string(item) and item.upper() == 'NONE'



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os


def _norm_path(path):
    if not path:
        return path
    return os.path.normpath(path.replace('/', os.sep))


def _compression_value_conversion(value):
    """
    PNG compression values are within range [0, 9]. This value must
    be mapped to a [0-100] interval.
    """
    try:
        if int(value) < 0 or int(value) > 100:
            raise RuntimeError("Quality argument must be of between 0 and 100.")
        return 0 if int(value) == 100 else int(9 - (int(value) / 11))
    except ValueError:
        raise RuntimeError("Quality argument must be of type integer.")


def _pil_quality_conversion(value):
    """
    The quality in Pillow is between [1, 95] and must be converted to
    a [0-100] interval.
    """
    try:
        if int(value) < 0 or int(value) > 100:
            raise RuntimeError("Quality argument must be of between 0 and 100.")
        if int(value) < 1:
            return 1
        elif int(value) >= 95:
            return 95
        return int(value)
    except ValueError:
        raise RuntimeError("The image quality argument must be of type integer.")


class suppress_stderr(object):

    def __init__(self):
        # Open a null file
        self.null_fd = os.open(os.devnull, os.O_RDWR)
        # Save the actual stderr (2) file descriptor.
        self.save_fd = os.dup(2)

    def __enter__(self):
        # Assign the null pointer to stderr.
        os.dup2(self.null_fd, 2)

    def __exit__(self, *_):
        # Re-assign the real stderr back to (2)
        os.dup2(self.save_fd, 2)
        # Close all file descriptors
        os.close(self.null_fd)
        os.close(self.save_fd)



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

import selenium

SeleniumVersion = namedtuple('SeleniumVersion', 'major minor micro')
major, minor, micro = (selenium.__version__.split('.') + ['0', '0'])[:3]
major, minor, micro = int(major), int(minor), int(micro)
SELENIUM_VERSION = SeleniumVersion(major=major, minor=minor, micro=micro)



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

from robot.libraries.DateTime import convert_date
from robot.utils import DotDict

from rbfWEB.base import LibraryComponent, keyword
from rbfWEB.errors import CookieNotFound
from rbfWEB.utils import is_truthy, is_noney, is_falsy


class CookieKeywords(LibraryComponent):

    @keyword
    def delete_all_cookies(self):
        """Deletes all cookies."""
        self.driver.delete_all_cookies()

    @keyword
    def delete_cookie(self, name):
        """Deletes cookie matching ``name``.

        If the cookie is not found, nothing happens.
        """
        self.driver.delete_cookie(name)

    @keyword
    def get_cookies(self, as_dict=False):
        """Returns all cookies of the current page.

        If ``as_dict`` argument evaluates as false, see `Boolean arguments` 
        for more details, then cookie information is returned as 
        a single string in format ``name1=value1; name2=value2; name3=value3``.
        When ``as_dict`` argument evaluates as true, cookie information
        is returned as Robot Framework dictionary format. The string format 
        can be used, for example, for logging purposes or in headers when
        sending HTTP requests. The dictionary format is helpful when
        the result can be passed to requests library's Create Session
        keyword's optional cookies parameter.
        
        
        """
        if is_falsy(as_dict):
            pairs = []
            for cookie in self.driver.get_cookies():
                pairs.append(cookie['name'] + "=" + cookie['value'])
            return '; '.join(pairs)
        else:
            pairs = DotDict()
            for cookie in self.driver.get_cookies():
                pairs[cookie['name']] = cookie['value']
            return pairs

    

    @keyword
    def get_cookie(self, name):
        """Returns information of cookie with ``name`` as an object.

        If no cookie is found with ``name``, keyword fails. The cookie object
        contains details about the cookie. Attributes available in the object
        are documented in the table below.

        | = Attribute = |             = Explanation =                                |
        | name          | The name of a cookie.                                      |
        | value         | Value of the cookie.                                       |
        | path          | Indicates a URL path, for example ``/``.                   |
        | domain        | The domain the cookie is visible to.                       |
        | secure        | When true, cookie is only used with HTTPS connections.     |
        | httpOnly      | When true, cookie is not accessible via JavaScript.        |
        | expiry        | Python datetime object indicating when the cookie expires. |
        | extra         | Possible attributes outside of the WebDriver specification |

        See the
        [https://w3c.github.io/webdriver/#cookies|WebDriver specification]
        for details about the cookie information.
        Notice that ``expiry`` is specified as a
        [https://docs.python.org/3/library/datetime.html#datetime.datetime|datetime object],
        not as seconds since Unix Epoch like WebDriver natively does.

        In some cases, example when running browser in the cloud, it is possible that
        cookie contains other attributes than is defined in the
        [https://w3c.github.io/webdriver/#cookies|WebDriver specification].
        These other attributes are available in a ``extra`` attribute in the cookie
        object and it contains a dictionary of the other attributes. 

        Example:
        | `Add Cookie`      | foo             | bar |
        | ${cookie} =       | `Get Cookie`    | foo |
        | `Should Be Equal` | ${cookie.name}  | bar |
        | `Should Be Equal` | ${cookie.value} | foo |
        | `Should Be True`  | ${cookie.expiry.year} > 2017 |

        
        """
        cookie = self.driver.get_cookie(name)
        if not cookie:
            raise CookieNotFound("Cookie with name '%s' not found." % name)
        return CookieInformation(**cookie)

    @keyword
    def add_cookie(self, name, value, path=None, domain=None, secure=None,
                   expiry=None):
        """Adds a cookie to your current session.

        ``name`` and ``value`` are required, ``path``, ``domain``, ``secure``
        and ``expiry`` are optional.  Expiry supports the same formats as
        the [http://robotframework.org/robotframework/latest/libraries/DateTime.html|DateTime]
        library or an epoch time stamp.

        Example:
        | `Add Cookie` | foo | bar |                            |
        | `Add Cookie` | foo | bar | domain=example.com         |
        | `Add Cookie` | foo | bar | expiry=2027-09-28 16:21:35 | # Expiry as timestamp.     |
        | `Add Cookie` | foo | bar | expiry=1822137695          | # Expiry as epoch seconds. |

        
        """
        new_cookie = {'name': name, 'value': value}
        if not is_noney(path):
            new_cookie['path'] = path
        if not is_noney(domain):
            new_cookie['domain'] = domain
        # Secure must be True or False
        if not is_noney(secure):
            new_cookie['secure'] = is_truthy(secure)
        if not is_noney(expiry):
            new_cookie['expiry'] = self._expiry(expiry)
        self.driver.add_cookie(new_cookie)

    def _expiry(self, expiry):
        try:
            return int(expiry)
        except ValueError:
            return int(convert_date(expiry, result_format='epoch'))


class CookieInformation(object):

    def __init__(self, name, value, path=None, domain=None, secure=False,
                 httpOnly=False, expiry=None, **extra):
        self.name = name
        self.value = value
        self.path = path
        self.domain = domain
        self.secure = secure
        self.httpOnly = httpOnly
        self.expiry = datetime.fromtimestamp(expiry) if expiry else None
        self.extra = extra

    def __str__(self):
        items = 'name value path domain secure httpOnly expiry'.split()
        string = '\n'.join('%s=%s' % (item, getattr(self, item))
                           for item in items)
        if self.extra:
            string = '%s%s=%s\n' % (string, 'extra', self.extra)
        return string



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import os.path
from copy import deepcopy
from .csvreader import CsvReader
from .xlsreader import XlsReader
from .pictreader import PictReader
from .version import __version__




class rbfDataDriver:
    """|
|



rbfDataDriver is a Data-Driven Testing library for Robot Framework.
This document explains how to use the rbfDataDriver library listener. 

rbfDataDriver is used/imported as Library but does not provide keywords
which can be used in a test. rbfDataDriver uses the Listener Interface
Version 3 to manipulate the test cases and creates new test cases based
on a Data-File that contains the data for Data-Driven Testing. These
data file may be .csv , .xls or .xlsx files.

rbfDataDriver is also able to cooperate with Microsoft PICT. An Open
Source Windows tool for data combination testing. Pict is able to
generate data combinations based on textual model definitions.
https://github.com/Microsoft/pict

|

Table of contents
---------------------

-  `What rbfDataDriver does <#WhatrbfDataDriverdoes>`__
-  `How rbfDataDriver works <#HowrbfDataDriverworks>`__
-  `Usage <#Usage>`__
-  `Structure of test suite <#Structureoftestsuite>`__
-  `Structure of data file <#Structureofdatafile>`__
-  `Data Sources <#DataSources>`__
-  `Encoding and CSV Dialect <#EncodingandCSVDialect>`__

|

What rbfDataDriver does
-------------------------

rbfDataDriver is an alternative approach to create Data-Driven Tests with
Robot Framework. rbfDataDriver creates multiple test cases based on a test
template and data content of a csv or Excel file. All created tests
share the same test sequence (keywords) and differ in the test data.
Because these tests are created on runtime only the template has to be
specified within the robot test specification and the used data are
specified in an external data file.

rbfDataDriver gives an alternative to the build in data driven approach
like:

.. code :: robotframework

    *** Settings ***
    Resource    login_resources.robot

    Suite Setup    Open my Browser
    Suite Teardown    Close Browsers
    Test Setup      Open Login Page
    Test Template    Invalid login


    *** Test Cases ***       User        Password
    Right user empty pass    demo        ${EMPTY}
    Right user wrong pass    demo        FooBar

    Empty user right pass    ${EMPTY}    mode
    Empty user empty pass    ${EMPTY}    ${EMPTY}
    Empty user wrong pass    ${EMPTY}    FooBar

    Wrong user right pass    FooBar      mode
    Wrong user empty pass    FooBar      ${EMPTY}
    Wrong user wrong pass    FooBar      FooBar

    *** Keywords ***
    Invalid login
        [Arguments]    ${username}    ${password}
        Input username    ${username}
        Input pwd    ${password}
        click login button
        Error page should be visible

This inbuild approach is fine for a hand full of data and a hand full of
test cases. If you have generated or calculated data and specially if
you have a variable amount of test case / combinations these robot files
becom quite a pain. With rbfDataDriver you may write the same test case
syntax but only once and deliver the data from en external data file.

One of the rare reasons when MicrosoftÂ® Excel or LibreOffice Calc may be
used in testingâ€¦ ;-)

`See example test suite <#example-suite>`__

`See example csv table <#example-csv>`__

|

How rbfDataDriver works
-------------------------

When the rbfDataDriver is used in a test suite it will be activated before
the test suite starts. It uses the Listener Interface Version 3 of Robot
Framework to read and modify the test specification objects. After
activation it searches for the ``Test Template`` -Keyword to analyze the
``[Arguments]`` it has. As a second step, it loads the data from the
specified CSV file. Based on the ``Test Template`` -Keyword, rbfDataDriver
creates as much test cases as lines are in the CSV file. As values for
the arguments of the ``Test Template`` -Keyword it reads values from the
column of the CSV file with the matching name of the ``[Arguments]``.
For each line of the CSV data table, one test case will be created. It
is also possible to specify test case names, tags and documentation for
each test case in the specific test suite related CSV file.

|

Usage
-------

rbfDataDriver is a â€œListenerâ€� but should not be set as a global listener
as command line option of robot. Because rbfDataDriver is a listener and a
library at the same time it sets itself as a listener when this library
is imported into a test suite.

To use it, just use it as Library in your suite. You may use the first
argument (option) which may set the file name or path to the data file.

Without any options set, it loads a .csv file which has the same name
and path like the test suite .robot .


**Example:**

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver

Options
~~~~~~~

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver
    ...    file=None,
    ...    encoding=cp1252
    ...    dialect=Excel-EU
    ...    delimiter=;
    ...    quotechar="
    ...    escapechar=\\\\
    ...    doublequote=True
    ...    skipinitialspace=False
    ...    lineterminator=\\r\\n


|

Encoding
^^^^^^^^

``encoding`` must be set if it shall not be cp1252

**cp1252** is the same like:

- Windows-1252
- Latin-1
- ANSI
- Windows Western European

See `Python Standard Encoding <https://docs.python.org/3/library/codecs.html#standard-encodings>`_ for more encodings

|

Example Excel (US / comma seperated)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Dialect Defaults:

.. code :: python

    delimiter = ','
    quotechar = '"'
    doublequote = True
    skipinitialspace = False
    lineterminator = '\\r\\n'
    quoting = QUOTE_MINIMAL

Usage in Robot Framework

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    my_data_file.csv    dialect=excel    encoding=${None}

|

Example Excel Tab (\\\\t seperated)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Dialect Defaults:

.. code :: python

    delimiter = '\\t'
    quotechar = '"'
    doublequote = True
    skipinitialspace = False
    lineterminator = '\\r\\n'
    quoting = QUOTE_MINIMAL

Usage in Robot Framework

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    my_data_file.csv    dialect=excel_tab

|

Example Unix Dialect
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Dialect Defaults:

.. code :: python

    delimiter = ','
    quotechar = '"'
    doublequote = True
    skipinitialspace = False
    lineterminator = '\\n'
    quoting = QUOTE_ALL

Usage in Robot Framework

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    my_data_file.csv    dialect=unix_dialect

|

Example User Defined
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

User may define the format completely free.
If an option is not set, the default values are used.
To register a userdefined format user have to set the
option ``dialect`` to ``UserDefined``


Usage in Robot Framework

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    my_data_file.csv
    ...    dialect=UserDefined
    ...    delimiter=.
    ...    lineterminator=\\n


|

Limitation
~~~~~~~~~~

|

Eclipse plug-in RED
^^^^^^^^^^^^^^^^^^^

There are known issues if the Eclipse plug-in RED is used. Because the
debugging Listener of this tool pre-calculates the number of test cases
before the creation of test cases by the rbfDataDriver. This leads to the
situation that the RED listener throws exceptions because it is called
for each test step but the RED GUI already stopped debugging so that the
listener cannot send Information to the GUI. This does not influence the
execution in any way but produces a lot of unwanted exceptions in the
Log.

|

Variable types
^^^^^^^^^^^^^^

In this early Version of rbfDataDriver, only scalar variables are
supported. Lists and dictionaries may be added in the next releases.

|

MS Excel and typed cells
^^^^^^^^^^^^^^^^^^^^^^^^

Microsoft Excel xls or xlsx file have the possibility to type thair data
cells. Numbers are typically of the type float. If these data are not
explicitly defined as text in Excel, pandas will read it as the type
that is has in excel. Because we have to work with strings in Robot
Framework these data are converted to string. This leads to the
situation that a European time value like â€œ04.02.2019â€� (4th January
2019) is handed over to Robot Framework in Iso time â€œ2019-01-04
00:00:00â€�. This may cause unwanted behavior. To mitigate this risk you
should define Excel based files explicitly as text within Excel.

|

How to activate the rbfDataDriver
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

To activate the rbfDataDriver for a test suite (one specific \*.robot file)
just import it as a library. You may also specify some options if the
default parameters do not fit your needs.

**Example**:

.. code :: robotframework

    *** Settings ***
    Library          rbfDataDriver
    Test Template    Invalid Logins

|

Structure of test suite
-----------------------

|

Requirements
~~~~~~~~~~~~

In the Moment there are some requirements how a test
suite must be structured so that the rbfDataDriver can get all the
information it needs.

 - only the first test case will be used as a template. All other test
   cases will be deleted.
 - Test cases have to be defined with a
   ``Test Template``. Reason for this is, that the rbfDataDriver needs to
   know the names of the test case arguments. Test cases do not have
   named arguments. Keywords do.
 - The keyword which is used as
   ``Test Template`` must be defined within the test suite (in the same
   \*.robot file). If the keyword which is used as ``Test Template`` is
   defined in a ``Resource`` the rbfDataDriver has no access to its
   arguments names.

|

Example Test Suite
~~~~~~~~~~~~~~~~~~

.. code :: robotframework

    ***Settings***
    Library           rbfDataDriver
    Resource          login_resources.robot
    Suite Setup       Open my Browser
    Suite Teardown    Close Browsers
    Test Setup        Open Login Page
    Test Template     Invalid Login

    *** Test Case ***
    Login with user ${username} and password ${password}    Default    UserData

    ***** *Keywords* *****
    Invalid login
        [Arguments]    ${username}    ${password}
        Input username    ${username}
        Input pwd    ${password}
        click login button
        Error page should be visible

In this example, the rbfDataDriver is activated by using it as a Library.
It is used with default settings.
As ``Test Template`` the keyword ``Invalid Login`` is used. This
keyword has two arguments. Argument names are ``${username}`` and
``${password}``. These names have to be in the CSV file as column
header. The test case has two variable names included in its name,
which does not have any functionality in Robot Framework. However, the
rbfDataDriver will use the test case name as a template name and
replaces the variables with the specific value of the single generated
test case.
This template test will only be used as a template. The specified data
``Default`` and ``UserData`` would only be used if no CSV file has
been found.

|

Structure of data file
----------------------

|

min. required columns
~~~~~~~~~~~~~~~~~~~~~

-  ``*** Test Cases ***`` column has to be the first one.
-  *Argument columns:* For each argument of the ``Test Template``
   keyword one column must be existing in the data file as data source.
   The name of this column must match the variable name and syntax.

|

optional columns
~~~~~~~~~~~~~~~~

-  *[Tags]* column may be used to add specific tags to a test case. Tags
   may be comma separated.
-  *[Documentation]* column may be used to add specific test case
   documentation.

|

Example Data file
~~~~~~~~~~~~~~~~~

+-------------+-------------+-------------+-------------+-------------+
| \**\* Test  | ${username} | ${password} | [Tags]      | [Documentat |
| Cases \**\* |             |             |             | ion]        |
|             |             |             |             |             |
+=============+=============+=============+=============+=============+
| Right user  | demo        | ${EMPTY}    | 1           | This is a   |
| empty pass  |             |             |             | test case   |
|             |             |             |             | documentati |
|             |             |             |             | on          |
|             |             |             |             | of the      |
|             |             |             |             | first one.  |
+-------------+-------------+-------------+-------------+-------------+
| Right user  | demo        | FooBar      | 2           |             |
| wrong pass  |             |             |             |             |
+-------------+-------------+-------------+-------------+-------------+
| empty user  | ${EMPTY}    | mode        | 1,2,3,4     | This test   |
| mode pass   |             |             |             | case has    |
|             |             |             |             | the Tags    |
|             |             |             |             | 1,2,3 and 4 |
|             |             |             |             | assigned.   |
+-------------+-------------+-------------+-------------+-------------+
|             | ${EMPTY}    | ${EMPTY}    |             | This test   |
|             |             |             |             | case has a  |
|             |             |             |             | generated   |
|             |             |             |             | name based  |
|             |             |             |             | on template |
|             |             |             |             | name.       |
+-------------+-------------+-------------+-------------+-------------+
|             | ${EMPTY}    | FooBar      |             | This test   |
|             |             |             |             | case has a  |
|             |             |             |             | generated   |
|             |             |             |             | name based  |
|             |             |             |             | on template |
|             |             |             |             | name.       |
+-------------+-------------+-------------+-------------+-------------+
|             | FooBar      | mode        |             | This test   |
|             |             |             |             | case has a  |
|             |             |             |             | generated   |
|             |             |             |             | name based  |
|             |             |             |             | on template |
|             |             |             |             | name.       |
+-------------+-------------+-------------+-------------+-------------+
|             | FooBar      | ${EMPTY}    |             | This test   |
|             |             |             |             | case has a  |
|             |             |             |             | generated   |
|             |             |             |             | name based  |
|             |             |             |             | on template |
|             |             |             |             | name.       |
+-------------+-------------+-------------+-------------+-------------+
|             | FooBar      | FooBar      |             | This test   |
|             |             |             |             | case has a  |
|             |             |             |             | generated   |
|             |             |             |             | name based  |
|             |             |             |             | on template |
|             |             |             |             | name.       |
+-------------+-------------+-------------+-------------+-------------+

In this data file, eight test cases are defined. Each line specifies one
test case. The first two test cases have specific names. The other six
test cases will generate names based on template test cases name with
the replacement of variables in this name. The order of columns is
irrelevant except the first column, ``*** Test Cases ***``

|

Data Sources
------------

|

CSV / TSV (Character-separated values)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

By default rbfDataDriver reads csv files. With the `Encoding and CSV
Dialect <#EncodingandCSVDialect>`__ settings you may configure which
structure your data source has.

|

XLS / XLSX Files
~~~~~~~~~~~~~~~~

If you want to use Excel based data sources, you may just set the file
to the extention or you may point to the correct file. If the extention
is â€œ.xlsâ€� or â€œ.xlsxâ€� rbfDataDriver will interpret it as Excel file. XLS
interpreter will ignore all other options like encoding, delimiters etc.

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    .xlsx

or:

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    file=my_data_source.xlsx

|

PICT (Pairwise Independent Combinatorial Testing)
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Pict is able to generate data files based on a model file.
https://github.com/Microsoft/pict

Documentation: https://github.com/Microsoft/pict/blob/master/doc/pict.md

|

Requirements
^^^^^^^^^^^^

-  Path to pict.exe must be set in the %PATH% environment variable.
-  Data model file has the file extention â€œ.pictâ€�
-  Pict model file must be encoded in UTF-8

|

How it works
^^^^^^^^^^^^

If the file option is set to a file with the extention pict, rbfDataDriver
will hand over this file to pict.exe and let it automatically generates
a file with the extention â€œ.pictoutâ€�. This file will the be used as data
source for the test generation. (It is tab seperated and UTF-8 encoded)
Except the file option all other options of the library will be ignored.

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    my_model_file.pict

|

CSV Encoding and CSV Dialect
----------------------------

CSV is far away from well designed and has absolutely no â€œcommonâ€�
format. Therefore it is possible to define your own dialect or use
predefined. The default is Excel-EU which is a semicolon separated
file.
These Settings are changeable as options of the rbfDataDriver Library.

|

file=
~~~~~

.. code :: robotframework

    *** Settings ***
    Library         rbfDataDriver    file=../data/my_data_source.csv


-  None(default): rbfDataDriver will search in the test suites folder if a
   \*.csv file with the same name than the test suite \*.robot file exists
-  only file extention: if you just set a file extentions like ".xls" or
   ".xlsx" rbfDataDriver will search
-  absolute path: If an absolute path to a file is set, rbfDataDriver tries
   to find and open the given data file.
-  relative path: If the option does not point to a data file as an
   absolute path, rbfDataDriver tries to find a data file relative to the
   folder where the test suite is located.

|

encoding=
~~~~~~~~~

may set the encoding of the CSV file. i.e.
``cp1252, ascii, iso-8859-1, latin-1, utf_8, utf_16, utf_16_be, utf_16_le``,
etcâ€¦ https://docs.python.org/3.7/library/codecs.html#standard-encodings

|

dialect=
~~~~~~~~

You may change the CSV Dialect here. If the Dialect is set to
â€˜UserDefinedâ€™ the following options are used. Otherwise, they are
ignored.
supported Dialects are:

.. code:: python

    "excel"
        delimiter = ','
        quotechar = '"'
        doublequote = True
        skipinitialspace = False
        lineterminator = '\\r\\n'
        quoting = QUOTE_MINIMAL

    "excel-tab"
        delimiter = '\\t'

    "unix"
        delimiter = ','
        quotechar = '"'
        doublequote = True
        skipinitialspace = False
        lineterminator = '\\n'
        quoting = QUOTE_ALL


Defaults:
~~~~~~~~~

.. code:: python

    file=None,
    encoding='cp1252',
    dialect='Excel-EU',
    delimiter=';',
    quotechar='"',
    escapechar='\\\\',
    doublequote=True,
    skipinitialspace=False,
    lineterminator='\\r\\n'

    """
    ROBOT_LIBRARY_DOC_FORMAT = 'reST'

    ROBOT_LISTENER_API_VERSION = 3
    ROBOT_LIBRARY_SCOPE = 'TEST SUITE'
    ROBOT_LIBRARY_VERSION = __version__

    TAGS_HEADER = '[Tags]'
    DOC_HEADER = '[Documentation]'

    def __init__(self,
                 file=None,
                 encoding='cp1252',
                 dialect='Excel-EU',
                 delimiter=';',
                 quotechar='"',
                 escapechar='\\',
                 doublequote=True,
                 skipinitialspace=False,
                 lineterminator='\r\n'
                 ):
        """**Example:**

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver

Options
~~~~~~~

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver
    ...    file=None,
    ...    encoding=cp1252
    ...    dialect=Excel-EU
    ...    delimiter=;
    ...    quotechar="
    ...    escapechar=\\\\
    ...    doublequote=True
    ...    skipinitialspace=False
    ...    lineterminator=\\r\\n


|

Encoding
^^^^^^^^

``encoding`` must be set if it shall not be cp1252

**cp1252** is the same like:

- Windows-1252
- Latin-1
- ANSI
- Windows Western European

See `Python Standard Encoding <https://docs.python.org/3/library/codecs.html#standard-encodings>`_ for more encodings

|

Example Excel (US / comma seperated)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Dialect Defaults:

.. code :: python

    delimiter = ','
    quotechar = '"'
    doublequote = True
    skipinitialspace = False
    lineterminator = '\\r\\n'
    quoting = QUOTE_MINIMAL

Usage in Robot Framework

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    my_data_file.csv    dialect=excel    encoding=${None}

|

Example Excel Tab (\\\\t seperated)
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Dialect Defaults:

.. code :: python

    delimiter = '\\t'
    quotechar = '"'
    doublequote = True
    skipinitialspace = False
    lineterminator = '\\r\\n'
    quoting = QUOTE_MINIMAL

Usage in Robot Framework

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    my_data_file.csv    dialect=excel_tab

|

Example Unix Dialect
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Dialect Defaults:

.. code :: python

    delimiter = ','
    quotechar = '"'
    doublequote = True
    skipinitialspace = False
    lineterminator = '\\n'
    quoting = QUOTE_ALL

Usage in Robot Framework

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    my_data_file.csv    dialect=unix_dialect

|

Example User Defined
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

User may define the format completely free.
If an option is not set, the default values are used.
To register a userdefined format user have to set the
option ``dialect`` to ``UserDefined``


Usage in Robot Framework

.. code :: robotframework

    *** Settings ***
    Library    rbfDataDriver    my_data_file.csv
    ...    dialect=UserDefined
    ...    delimiter=.
    ...    lineterminator=\\n

        """
        self.ROBOT_LIBRARY_LISTENER = self

        self.file = file
        self.csv_encoding = encoding
        self.csv_dialect = dialect
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.escapechar = escapechar
        self.doublequote = doublequote
        self.skipinitialspace = skipinitialspace
        self.lineterminator = lineterminator

        self.testcase_table_name = '*** Test Cases ***'

        self.suite_source = None
        self.template_test = None
        self.template_keyword = None
        self.data_table = None
        self.index = None

    # noinspection PyUnusedLocal
    def _start_suite(self, suite, result):
        """Called when a test suite starts.
        Data and result are model objects representing the executed test suite and its execution results, respectively.

        :param suite: class robot.running.model.TestSuite(name='', doc='', metadata=None, source=None)
        :param result: NOT USED
        """
        try:
            if suite.rpa:
                self.testcase_table_name = '*** Tasks ***'
        except AttributeError:
            pass

        self.suite_source = suite.source
        self._create_data_table()
        self.template_test = suite.tests[0]
        self.template_keyword = self._get_template_keyword(suite)
        temp_test_list = list()
        for self.index, lines in enumerate(self.data_table[self.testcase_table_name]):
            self._create_test_from_template()
            temp_test_list.append(self.test)
        suite.tests = temp_test_list

    def _create_data_table(self):
        """
        this function creates a dictionary which contains all data from data file.
        Keys are header names.
        Values are data of this column as array.
        """

        if (not self.file) or ('' == self.file[:self.file.rfind('.')]):
            self._get_data_file_from_suite_source()
        else:
            self._check_if_file_exists_as_path_or_in_suite()

        filename, file_extension = os.path.splitext(self.file)
        if (file_extension.lower() == '.xlsx') or (file_extension == '.xls'):
            xls = XlsReader(self.file)
            self.data_table = xls.get_data_from_xls()

        elif file_extension.lower() == '.pict':
            pict = PictReader(self.file, self.testcase_table_name)
            self.data_table = pict.get_data_from_pict()
        else:
            csv = CsvReader(self.file,
                            self.csv_encoding,
                            self.testcase_table_name,
                            self.csv_dialect,
                            self.delimiter,
                            self.quotechar,
                            self.escapechar,
                            self.doublequote,
                            self.skipinitialspace,
                            self.lineterminator)

            self.data_table = csv.get_data_from_csv()

    def _get_data_file_from_suite_source(self):
        if not self.file:
            suite_path_as_data_file = f'{self.suite_source[:self.suite_source.rfind(".")]}.csv'
        else:
            suite_path_as_data_file = \
                f'{self.suite_source[:self.suite_source.rfind(".")]}{self.file[self.file.rfind("."):]}'

        if os.path.isfile(suite_path_as_data_file):
            self.file = suite_path_as_data_file
        else:
            raise FileNotFoundError(
                'File attribute was empty. Tried to find '
                + suite_path_as_data_file + ' but file does not exist.')

    def _check_if_file_exists_as_path_or_in_suite(self):
        if not os.path.isfile(self.file):
            suite_dir = str(os.path.dirname(self.suite_source))
            file_in_suite_dir = os.path.join(suite_dir, self.file)
            if os.path.isfile(file_in_suite_dir):
                self.file = file_in_suite_dir
            else:
                raise FileNotFoundError(
                    'File attribute was not a full path. Tried to find '
                    + file_in_suite_dir + ' but file does not exist.')

    def _get_template_keyword(self, suite):
        template = self.template_test.template
        if template:
            for keyword in suite.resource.keywords:
                if self._is_same_keyword(keyword.name, template):
                    return keyword
        raise AttributeError('No "Test Template" keyword found for first test case.')

    # noinspection PyMethodMayBeStatic
    def _get_normalized_keyword(self, keyword):
        return keyword.lower().replace(' ', '').replace('_', '')

    def _is_same_keyword(self, first, second):
        return self._get_normalized_keyword(first) == self._get_normalized_keyword(second)

    def _create_test_from_template(self):
        self.test = deepcopy(self.template_test)
        self._replace_test_case_name()
        self._replace_test_case_keywords()
        self._add_test_case_tags()
        self._replace_test_case_doc()

    def _add_test_case_tags(self):
        if self.TAGS_HEADER in self.data_table:
            for tag in self.data_table[self.TAGS_HEADER][self.index].split(','):
                self.test.tags.add(tag.strip())

    def _replace_test_case_doc(self):
        if self.DOC_HEADER in self.data_table:
            self.test.doc = self.data_table[self.DOC_HEADER][self.index]

    def _replace_test_case_keywords(self):
        self.test.keywords.clear()
        if self.template_test.keywords.setup is not None:
            self.test.keywords.create(name=self.template_test.keywords.setup.name, type='setup',
                                      args=self.template_test.keywords.setup.args)
        self.test.keywords.create(name=self.template_keyword.name,
                                  args=self._get_template_args_for_index())
        if self.template_test.keywords.teardown is not None:
            self.test.keywords.create(name=self.template_test.keywords.teardown.name, type='teardown',
                                      args=self.template_test.keywords.teardown.args)

    def _replace_test_case_name(self):
        if self.data_table[self.testcase_table_name][self.index] == '':
            for key in self.data_table.keys():
                if key[:1] == '$':
                    self.test.name = self.test.name.replace(key, self.data_table[key][self.index])
        else:
            self.test.name = self.data_table[self.testcase_table_name][self.index]

    def _get_template_args_for_index(self):
        return_args = []
        for arg in self.template_keyword.args:
            if arg in self.data_table:
                return_args.append(self.data_table[arg][self.index])
            else:
                return_args.append(arg)
        return return_args



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import types

from selenium import webdriver

from rbfWEB.base import keyword, LibraryComponent
from rbfWEB.locators import WindowManager
from rbfWEB.utils import (is_truthy, is_noney, secs_to_timestr,
                                   timestr_to_secs)

from .webdrivertools import WebDriverCreator


class BrowserManagementKeywords(LibraryComponent):

    def __init__(self, ctx):
        LibraryComponent.__init__(self, ctx)
        self._window_manager = WindowManager(ctx)

    @keyword
    def close_all_browsers(self):
        """Closes all open browsers and resets the browser cache.

        After this keyword new indexes returned from `Open Browser` keyword
        are reset to 1.

        This keyword should be used in test or suite teardown to make sure
        all browsers are closed.
        """
        self.debug('Closing all browsers.')
        self.drivers.close_all()

    @keyword
    def close_browser(self):
        """Closes the current browser."""
        if self.drivers.current:
            self.debug('Closing browser with session id {}.'
                       .format(self.driver.session_id))
            self.drivers.close()

    @keyword
    def open_browser(self, url, browser='firefox', alias=None,
                     remote_url=False, desired_capabilities=None,
                     ff_profile_dir=None):
        """Opens a new browser instance to the given ``url``.

        The ``browser`` argument specifies which browser to use, and the
        supported browser are listed in the table below. The browser names
        are case-insensitive and some browsers have multiple supported names.

        |    = Browser =    |        = Name(s) =       |
        | Firefox           | firefox, ff              |
        | Google Chrome     | googlechrome, chrome, gc |
        | Headless Firefox  | headlessfirefox          |
        | Headless Chrome   | headlesschrome           |
        | Internet Explorer | internetexplorer, ie     |
        | Edge              | edge                     |
        | Safari            | safari                   |
        | Opera             | opera                    |
        | Android           | android                  |
        | Iphone            | iphone                   |
        | PhantomJS         | phantomjs                |
        | HTMLUnit          | htmlunit                 |
        | HTMLUnit with Javascript | htmlunitwithjs    |

        To be able to actually use one of these browsers, you need to have
        a matching Selenium browser driver available. Headless Firefox and
        Headless Chrome require Selenium 3.8.0 or newer.

        Optional ``alias`` is an alias given for this browser instance and
        it can be used for switching between browsers. An alternative
        approach for switching is using an index returned by this keyword.
        These indices start from 1, are incremented when new browsers are
        opened, and reset back to 1 when `Close All Browsers` is called.
        See `Switch Browser` for more information and examples.

        Optional ``remote_url`` is the URL for a
        [https://github.com/SeleniumHQ/selenium/wiki/Grid2|Selenium Grid].

        Optional ``desired_capabilities`` can be used to configure, for example,
        logging preferences for a browser or a browser and operating system
        when using [http://saucelabs.com|Sauce Labs]. Desired capabilities can
        be given either as a Python dictionary or as a string in format
        ``key1:value1,key2:value2``.
        [https://github.com/SeleniumHQ/selenium/wiki/DesiredCapabilities|
        Selenium documentation] lists possible capabilities that can be
        enabled.

        Optional ``ff_profile_dir`` is the path to the Firefox profile
        directory if you wish to overwrite the default profile Selenium
        uses. 

        Examples:
        | `Open Browser` | http://example.com | Chrome  |
        | `Open Browser` | http://example.com | Firefox | alias=Firefox |
        | `Open Browser` | http://example.com | Edge    | remote_url=http://127.0.0.1:4444/wd/hub |

        If the provided configuration options are not enough, it is possible
        to use `Create Webdriver` to customize browser initialization even
        more.

        Applying ``desired_capabilities`` argument also for local browser is
        
        """
        if is_truthy(remote_url):
            self.info("Opening browser '%s' to base url '%s' through "
                      "remote server at '%s'." % (browser, url, remote_url))
        else:
            self.info("Opening browser '%s' to base url '%s'." % (browser, url))
        driver = self._make_driver(browser, desired_capabilities,
                                   ff_profile_dir, remote_url)
        try:
            driver.get(url)
        except Exception:
            self.ctx.register_driver(driver, alias)
            self.debug("Opened browser with session id %s but failed "
                       "to open url '%s'." % (driver.session_id, url))
            raise
        self.debug('Opened browser with session id %s.' % driver.session_id)
        return self.ctx.register_driver(driver, alias)

    @keyword
    def create_webdriver(self, driver_name, alias=None, kwargs={},
                         **init_kwargs):
        """Creates an instance of Selenium WebDriver.

        Like `Open Browser`, but allows passing arguments to the created
        WebDriver instance directly. This keyword should only be used if
        functionality provided by `Open Browser` is not adequate.

        ``driver_name`` must be an WebDriver implementation name like Firefox,
        Chrome, Ie, Opera, Safari, PhantomJS, or Remote.

        The initialized WebDriver can be configured either with a Python
        dictionary ``kwargs`` or by using keyword arguments ``**init_kwargs``.
        These arguments are passed directly to WebDriver without any
        processing. See [https://seleniumhq.github.io/selenium/docs/api/py/api.html|
        Selenium API documentation] for details about the supported arguments.

        Examples:
        | # Use proxy with Firefox   |                |                                           |                         |
        | ${proxy}=                  | `Evaluate`     | sys.modules['selenium.webdriver'].Proxy() | sys, selenium.webdriver |
        | ${proxy.http_proxy}=       | `Set Variable` | localhost:8888                            |                         |
        | `Create Webdriver`         | Firefox        | proxy=${proxy}                            |                         |
        | # Use proxy with PhantomJS |                |                                           |                         |
        | ${service args}=           | `Create List`  | --proxy=192.168.132.104:8888              |                         |
        | `Create Webdriver`         | PhantomJS      | service_args=${service args}              |                         |

        Returns the index of this browser instance which can be used later to
        switch back to it. Index starts from 1 and is reset back to it when
        `Close All Browsers` keyword is used. See `Switch Browser` for an
        example.
        """
        if not isinstance(kwargs, dict):
            raise RuntimeError("kwargs must be a dictionary.")
        for arg_name in kwargs:
            if arg_name in init_kwargs:
                raise RuntimeError("Got multiple values for argument '%s'." % arg_name)
            init_kwargs[arg_name] = kwargs[arg_name]
        driver_name = driver_name.strip()
        try:
            creation_func = getattr(webdriver, driver_name)
        except AttributeError:
            raise RuntimeError("'%s' is not a valid WebDriver name." % driver_name)
        self.info("Creating an instance of the %s WebDriver." % driver_name)
        driver = creation_func(**init_kwargs)
        self.debug("Created %s WebDriver instance with session id %s."
                   % (driver_name, driver.session_id))
        return self.ctx.register_driver(driver, alias)

    @keyword
    def switch_browser(self, index_or_alias):
        """Switches between active browsers using ``index_or_alias``.

        Indices are returned by the `Open Browser` keyword and aliases can
        be given to it explicitly. Indices start from 1.

        Example:
        | `Open Browser`        | http://google.com | ff       |
        | `Location Should Be`  | http://google.com |          |
        | `Open Browser`        | http://yahoo.com  | ie       | alias=second |
        | `Location Should Be`  | http://yahoo.com  |          |
        | `Switch Browser`      | 1                 | # index  |
        | `Page Should Contain` | I'm feeling lucky |          |
        | `Switch Browser`      | second            | # alias  |
        | `Page Should Contain` | More Yahoo!       |          |
        | `Close All Browsers`  |                   |          |

        Above example expects that there was no other open browsers when
        opening the first one because it used index ``1`` when switching to
        it later. If you are not sure about that, you can store the index
        into a variable as below.

        | ${index} =         | `Open Browser` | http://google.com |
        | # Do something ... |                |                   |
        | `Switch Browser`   | ${index}       |                   |
        """
        try:
            self.drivers.switch(index_or_alias)
        except RuntimeError:
            raise RuntimeError("No browser with index or alias '%s' found."
                               % index_or_alias)
        self.debug('Switched to browser with Selenium session id %s.'
                   % self.driver.session_id)

    @keyword
    def get_session_id(self):
        """Returns the currently active browser session id.

        
        """
        return self.driver.session_id

    @keyword
    def get_source(self):
        """Returns the entire HTML source of the current page or frame."""
        return self.driver.page_source

    @keyword
    def get_title(self):
        """Returns the title of current page."""
        return self.driver.title

    @keyword
    def get_location(self):
        """Returns the current browser URL."""
        return self.driver.current_url

    @keyword
    def location_should_be(self, url, message=None):
        """Verifies that current URL is exactly ``url``.

        The ``url`` argument contains the exact url that should exist in browser.

        The ``message`` argument can be used to override the default error
        message.

        
        """
        actual = self.get_location()
        if actual != url:
            if is_noney(message):
                message = ("Location should have been '%s' but "
                           "was '%s'." % (url, actual))
            raise AssertionError(message)
        self.info("Current location is '%s'." % url)

    @keyword
    def location_should_contain(self, expected, message=None):
        """Verifies that current URL contains ``expected``.

        The ``expected`` argument contains the expected value in url.

        The ``message`` argument can be used to override the default error
        message.

        
        """
        actual = self.get_location()
        if expected not in actual:
            if is_noney(message):
                message = ("Location should have contained '%s' but "
                           "it was '%s'." % (expected, actual))
            raise AssertionError(message)
        self.info("Current location contains '%s'." % expected)

    @keyword
    def log_location(self):
        """Logs and returns the current URL."""
        url = self.get_location()
        self.info(url)
        return url

    @keyword
    def log_source(self, loglevel='INFO'):
        """Logs and returns the HTML source of the current page or frame.

        The ``loglevel`` argument defines the used log level. Valid log
        levels are ``WARN``, ``INFO`` (default), ``DEBUG``, ``TRACE``
        and ``NONE`` (no logging).
        """
        source = self.get_source()
        self.log(source, loglevel)
        return source

    @keyword
    def log_title(self):
        """Logs and returns the title of current page."""
        title = self.get_title()
        self.info(title)
        return title

    @keyword
    def title_should_be(self, title, message=None):
        """Verifies that current page title equals ``title``.

        The ``message`` argument can be used to override the default error
        message.

        
        """
        actual = self.get_title()
        if actual != title:
            if is_noney(message):
                message = "Title should have been '%s' but was '%s'." % (title, actual)
            raise AssertionError(message)
        self.info("Page title is '%s'." % title)

    @keyword
    def go_back(self):
        """Simulates the user clicking the back button on their browser."""
        self.driver.back()

    @keyword
    def go_to(self, url):
        """Navigates the active browser instance to the provided ``url``."""
        self.info("Opening url '%s'" % url)
        self.driver.get(url)

    @keyword
    def reload_page(self):
        """Simulates user reloading page."""
        self.driver.refresh()

    @keyword
    def get_selenium_speed(self):
        """Gets the delay that is waited after each Selenium command.

        The value is returned as a human readable string like ``1 second``.

        See the `Selenium Speed` section above for more information.
        """
        return secs_to_timestr(self.ctx.speed)

    @keyword
    def get_selenium_timeout(self):
        """Gets the timeout that is used by various keywords.

        The value is returned as a human readable string like ``1 second``.

        See the `Timeout` section above for more information.
        """
        return secs_to_timestr(self.ctx.timeout)

    @keyword
    def get_selenium_implicit_wait(self):
        """Gets the implicit wait value used by Selenium.

        The value is returned as a human readable string like ``1 second``.

        See the `Implicit wait` section above for more information.
        """
        return secs_to_timestr(self.ctx.implicit_wait)

    @keyword
    def set_selenium_speed(self, value):
        """Sets the delay that is waited after each Selenium command.

        The value can be given as a number that is considered to be
        seconds or as a human readable string like ``1 second``.
        The previous value is returned and can be used to restore
        the original value later if needed.

        See the `Selenium Speed` section above for more information.

        Example:
        | `Set Selenium Speed` | 0.5 seconds |
        """
        old_speed = self.get_selenium_speed()
        self.ctx.speed = timestr_to_secs(value)
        for driver in self.drivers.active_drivers:
            self._monkey_patch_speed(driver)
        return old_speed

    @keyword
    def set_selenium_timeout(self, value):
        """Sets the timeout that is used by various keywords.

        The value can be given as a number that is considered to be
        seconds or as a human readable string like ``1 second``.
        The previous value is returned and can be used to restore
        the original value later if needed.

        See the `Timeout` section above for more information.

        Example:
        | ${orig timeout} = | `Set Selenium Timeout` | 15 seconds |
        | `Open page that loads slowly` |
        | `Set Selenium Timeout` | ${orig timeout} |
        """
        old_timeout = self.get_selenium_timeout()
        self.ctx.timeout = timestr_to_secs(value)
        for driver in self.drivers.active_drivers:
            driver.set_script_timeout(self.ctx.timeout)
        return old_timeout

    @keyword
    def set_selenium_implicit_wait(self, value):
        """Sets the implicit wait value used by Selenium.

        The value can be given as a number that is considered to be
        seconds or as a human readable string like ``1 second``.
        The previous value is returned and can be used to restore
        the original value later if needed.

        This keyword sets the implicit wait for all opened browsers.
        Use `Set Browser Implicit Wait` to set it only to the current
        browser.

        See the `Implicit wait` section above for more information.

        Example:
        | ${orig wait} = | `Set Selenium Implicit Wait` | 10 seconds |
        | `Perform AJAX call that is slow` |
        | `Set Selenium Implicit Wait` | ${orig wait} |
        """
        old_wait = self.get_selenium_implicit_wait()
        self.ctx.implicit_wait = timestr_to_secs(value)
        for driver in self.drivers.active_drivers:
            driver.implicitly_wait(self.ctx.implicit_wait)
        return old_wait

    @keyword
    def set_browser_implicit_wait(self, value):
        """Sets the implicit wait value used by Selenium.

        Same as `Set Selenium Implicit Wait` but only affects the current
        browser.
        """
        self.driver.implicitly_wait(timestr_to_secs(value))

    def _make_driver(self, browser, desired_capabilities=None,
                     profile_dir=None, remote=None):
        driver = WebDriverCreator(self.log_dir).create_driver(
            browser=browser, desired_capabilities=desired_capabilities,
            remote_url=remote, profile_dir=profile_dir)
        driver.set_script_timeout(self.ctx.timeout)
        driver.implicitly_wait(self.ctx.implicit_wait)
        if self.ctx.speed:
            self._monkey_patch_speed(driver)
        return driver

    def _monkey_patch_speed(self, driver):
        def execute(self, driver_command, params=None):
            result = self._base_execute(driver_command, params)
            speed = self._speed if hasattr(self, '_speed') else 0.0
            if speed > 0:
                time.sleep(speed)
            return result
        if not hasattr(driver, '_base_execute'):
            driver._base_execute = driver.execute
            driver.execute = types.MethodType(execute, driver)
        driver._speed = self.ctx.speed



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .version import VERSION
from .client import Client

__version__ = VERSION


class rbfVideoRecorder:
    """ Test Library for taking screenshots on the machine where tests are run.

    Note that successfully taking screenshots requires tests to be run with
    a physical or virtual display.

    = Usage =

    How screenshots are taken when using Python depends on the chosen format.
    There are several supported formats: PNG, JPG/JPEG, GIF and WebP. If no format
    is specified the screenshots will have the PNG format.

    For taking screenshots the following modules are used:

    - [https://python-mss.readthedocs.io| mss ] a fast cross-platform module used for taking screenshots and saving
     them in PNG format.

    - [https://pillow.readthedocs.io | Pillow] used on top of ``mss`` in order to save the screenshots in JPG/JPEG/WebP format.

    - [http://pygtk.org/ | PyGTK] is an alternative to ``mss`` for taking screenshots when using VNC.


    For video recording, [https://github.com/skvark/opencv-python/blob/master/README.md | OpenCV-Python] is used and
    the output file is in WebM format.

    = Where screenshots are saved =

    By default screenshots are saved into the same directory where the Robot
    Framework log file is written. If no log is created, screenshots are
    saved into the directory where the XML output file is written.

    It is possible to specify a custom location for screenshots using
    ``screenshot_directory`` argument when `importing` the library and using
    `Set Screenshot Directory` keyword during execution. It is also possible
    to save screenshots using an absolute path.

    = Time format =

    All delays and time intervals can be given as numbers considered seconds
    (e.g. ``0.5`` or ``42``) or in Robot Framework's time syntax
    (e.g. ``1.5 seconds`` or ``1 min 30 s``). For more information about
    the time syntax see the
    [http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#time-format|Robot Framework User Guide].

    = Boolean arguments =

    Some keywords accept arguments that are handled as Boolean values true or
    false. If such an argument is given as a string, it is considered false if
    it is either an empty string or case-insensitively equal to ``false``,
    ``none`` or ``no``. Other strings are considered true regardless
    their value, and other argument types are tested using the same
    [http://docs.python.org/2/library/stdtypes.html#truth-value-testing|rules
    as in Python].

    True examples:
    | `Take Partial Screenshot` | embed=True    | # Strings are generally true.    |
    | `Take Partial Screenshot` | embed=yes     | # Same as the above.             |
    | `Take Partial Screenshot` | embed=${TRUE} | # Python ``True`` is true.       |
    | `Take Partial Screenshot` | embed=${42}   | # Numbers other than 0 are true. |
    False examples:
    | `Take Partial Screenshot` | embed=False    | # String ``false`` is false.   |
    | `Take Partial Screenshot` | embed=no       | # Also string ``no`` is false. |
    | `Take Partial Screenshot` | embed=${EMPTY} | # Empty string is false.       |
    | `Take Partial Screenshot` | embed=${FALSE} | # Python ``False`` is false.   |
    """

    ROBOT_LIBRARY_VERSION = __version__

    def __init__(self, screenshot_module=None, screenshot_directory=None, format='png', quality=50, delay=0, fps=8):
        """
        ``screenshot_module`` specifies the module or tool to use when taking screenshots using this library.
        If no tool or module is specified, ``mss`` will be used by default. For running
        on Linux with VNC, use ``PyGTK``.

        To configure where screenshots are saved use ``screenshot_directory``. If no value is given,
        screenshots are saved into same directory as the log file. The directory can also be set using
        `Set Screenshot Directory` keyword.

        ``format`` specifies the format in which the screenshots will be saved.
        Possible values are ``png``,  ``jpeg``, ``jpg`` and ``webp``, case-insensitively.
        If no value is given the format is set by default to ``png``.

        ``quality`` can take values in range [0, 100]. Value 0 is lowest quality,
        while value 100 is maximum quality. The quality is directly proportional
        with file size. Because PNG uses lossless compression its size
        may be larger than the size of the JPG file. The default value is 50.

        ``delay`` specifies the waiting time before taking a screenshot. See
        `Time format` section for more information. By default the delay is 0.

        Examples (use only one of these):
        | =Setting= |  =Value=   |  =Value=                        |
        | Library   | Screenshot |                                 |
        | Library   | Screenshot | PyGTK                           |
        | Library   | Screenshot | screenshot_directory=${TEMPDIR} |
        | Library   | Screenshot | format=jpg                      |
        | Library   | Screenshot | quality=0                       |

        """
        self._client = Client(
            screenshot_module=screenshot_module,
            screenshot_directory=screenshot_directory,
            format=format,
            quality=quality,
            delay=delay,
            fps=fps
        )

    def set_screenshot_directory(self, path):
        """Sets the directory where screenshots are saved.

        It is possible to use ``/`` as a path separator in all operating
        systems. Path to the old directory is returned.

        The directory can also be set in `importing`.
        """
        return self._client.set_screenshot_directory(path)

    def take_screenshot(self, name='screenshot', format=None, quality=None, width='800px', delay=0):
        """Takes a screenshot in the specified format at library import and
        embeds it into the log file (PNG by default).

        Name of the file where the screenshot is stored is derived from the
        given ``name``. If the ``name`` ends with extension ``.jpg``, ``.jpeg``,
        ``.png`` or ``.webp``, the screenshot will be stored with that exact name.
        Otherwise a unique name is created by adding an underscore, a running
        index and an extension to the ``name``.

        The name will be interpreted to be relative to the directory where
        the log file is written. It is also possible to use absolute paths.
        Using ``/`` as a path separator works in all operating systems.

        ``format`` specifies the format in which the screenshot is saved. If
        no format is provided the library import value will be used which is
        ``png`` by default. Can be either ``jpg``, ``jpeg``, ``png``, or ``webp``,
        case insensitive.

        ``quality`` can take values in range [0, 100]. In case of JPEG format
        it can drastically reduce the file size of the image.

        ``width`` specifies the size of the screenshot in the log file.

        ``delay`` specifies the waiting time before taking a screenshot. See
        `Time format` section for more information. By default the delay is 0.

        Examples: (LOGDIR is determined automatically by the library)
        | `Take Screenshot` |                  |            | # LOGDIR/screenshot_1.png (index automatically incremented) |
        | `Take Screenshot` | mypic            |            | # LOGDIR/mypic_1.png (index automatically incremented) |
        | `Take Screenshot` | ${TEMPDIR}/mypic |            | # /tmp/mypic_1.png (index automatically incremented) |
        | `Take Screenshot` | pic.jpg          |            | # LOGDIR/pic.jpg (always uses this file) |
        | `Take Screenshot` | images/login.jpg | 300px      | # Specify both name and width. |
        | `Take Screenshot` | width=550px      |            | # Specify only width. |
        | `Take Screenshot` | format=jpg       | quality=15 | # Specify both image format and quality |

        The path where the screenshot is saved is returned.
        """
        return self._client.take_screenshot(name, format, quality, width, delay)

    def start_gif_recording(self, name="screenshot", size_percentage=0.5,
                            embed=True, embed_width='800px'):
        """
        Starts the recording of a GIF in the background with the specified ``name``.
        The recording can be stopped by calling the `Stop Gif Recording` keyword.

        ``name`` specifies the name by which the screenshot will be saved.

        ``size_percentage`` is used to reduce the size of the GIFs a resize of the
        screencaptures. It will specify how much this reduction is with respect to
        screen resolution. By default this parameter is set to resize the images to
        0.5 of the screen resolution.

        ``embed`` specifies if the screenshot should be embedded in the log file
        or not. See `Boolean arguments` section for more details.

        ``embed_width`` specifies the size of the screenshot that is
        embedded in the log file.

        Examples:
        | `Start Gif Recording` |            |  # Starts the GIF recording in background |
        | `Sleep`               | 10 seconds |  # Here should be the actions that will be recorded |
        | `Stop Gif Recording`  |            |  # Will create the GIF containing the screen recording \
        since `Start Gif Recording` was called. |

        *Note:* Recommended GIF duration is 100 seconds or lower depending on system memory.
        """
        return self._client.start_gif_recording(name, size_percentage, embed, embed_width)

    def stop_gif_recording(self):
        """
        Stops the GIF recording and generates the file. If ``embed`` argument was set to ``True`` the
        GIF will be displayed in the log file.
        """
        self._client.stop_gif_recording()

    def take_partial_screenshot(self, name='screenshot', format=None, quality=None,
                                left=0, top=0, width=700, height=300, embed=True, embed_width='800px'):
        """
        Takes a partial screenshot in the specified format and dimensions at
        library import and embeds it into the log file (PNG by default).

        This keyword is similar with `Take Screenshot` but has some extra parameters listed below:.

        ``left`` specifies the cropping distance on the X axis from the left of the screen capture.

        ``top`` specifies the cropping distance on the Y axis from the top of the screen capture.

        ``width`` specifies the width of a screen capture when using partial screen captures.

        ``height`` specifies the height of a screen capture when using partial screen captures.

        ``embed`` specifies if the screenshot should be embedded in the log file or not. See
        `Boolean arguments` section for more details.

        ``embed_width`` specifies the size of the screenshot that is embedded in the log file.
        """
        return self._client.take_partial_screenshot(name, format, quality,
                                                    left, top, width, height, embed, embed_width)

    def take_screenshot_without_embedding(self, name="screenshot", format=None, quality=None, delay=0):
        """Takes a screenshot and links it from the log file.
        This keyword is otherwise identical to `Take Screenshot` but the saved
        screenshot is not embedded into the log file. The screenshot is linked
        so it is nevertheless easily available.
        """
        return self._client.take_screenshot_without_embedding(name, format, quality, delay)

    def take_multiple_screenshots(self, name="screenshot", format=None, quality=None,
                                  screenshot_number=2, delay_time=0):
        """Takes the specified number of screenshots in the specified format
        at library import and embeds it into the log file (PNG by default).

        This keyword is similar with `Take Screenshot` but has some extra
        parameters listed below:

        ``screenshot_number`` specifies the number of screenshots to be taken.
        By default this number is 2.

        ``delay_time`` specifies the waiting time before taking another
        screenshot. See `Time format` section for more information. By
        default the delay time is 0.
        """
        return self._client.take_multiple_screenshots(name, format, quality, screenshot_number, delay_time)

    def start_video_recording(self, name="recording", fps=8, embed=True, embed_width='800px'):
        """Starts the recording of a video in the background with the specified ``name``.
        The recording can be stopped by calling the `Stop Video Recording` keyword.

        ``name`` specifies the name by which the record will be saved.

        ``fps`` specifies the frame rate at which the video is displayed.
        By default frame rate is 8.

        ``embed`` specifies if the record should be embedded in the log file
        or not. See `Boolean arguments` section for more details.

        ``embed_width`` specifies the size of the video record that is
        embedded in the log file.

        Examples:
        | `Start Video Recording` |            |  # Starts the video recording in background |
        | `Sleep`                 | 10 seconds |  # Here should be the actions that will be recorded |
        | `Stop Video Recording`  |            |  # Will create the video containing the screen recording \
        since `Start Video Recording` was called. |

        *Note:* Be aware that during recording the number of the collected frames are dependent on the
        performance of your system. Check different values for ``fps`` to find optimal results. (e.g.
        for a 15 seconds recording the output might be a 2 second video with 30 fps).
        """
        return self._client.start_video_recording(name, fps, embed, embed_width)

    def stop_video_recording(self):
        """Stops the video recording and generates the file in WebM format. If ``embed`` argument
        was set to ``True`` the video will be displayed in the log file.
        """
        self._client.stop_video_recording()



# Copyright 2019-     DNB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

from rbfWEB.utils import is_falsy, timestr_to_secs
from selenium.common.exceptions import NoSuchWindowException

from rbfWEB.base import keyword, LibraryComponent
from rbfWEB.locators import WindowManager
from rbfWEB.utils import plural_or_not


class WindowKeywords(LibraryComponent):

    def __init__(self, ctx):
        LibraryComponent.__init__(self, ctx)
        self._window_manager = WindowManager(ctx)

    @keyword
    def select_window(self, locator='MAIN', timeout=None):
        """Selects browser window matching ``locator``.

        If the window is found, all subsequent commands use the selected
        window, until this keyword is used again. If the window is not
        found, this keyword fails. The previous window handle is returned,
        and can be used to return back to it later.

        Notice that in this context _window_ means a pop-up window opened
        when doing something on an existing window. It is not possible to
        select windows opened with `Open Browser`, `Switch Browser` must
        be used instead. Notice also that alerts should be handled with
        `Handle Alert` or other alert related keywords.

        The ``locator`` can be specified using different strategies somewhat
        similarly as when `locating elements` on pages.

        - By default the ``locator`` is matched against window handle, name,
          title, and URL. Matching is done in that order and the the first
          matching window is selected.

        - The ``locator`` can specify an explicit strategy by using format
          ``strategy:value`` (recommended) or ``strategy=value``. Supported
          strategies are ``name``, ``title`` and ``url``, which match windows
          using name, title, and URL, respectively. Additionally, ``default``
          can be used to explicitly use the default strategy explained above.

        - If the ``locator`` is ``NEW`` (case-insensitive), the latest
          opened window is selected. It is an error if this is the same
          as the current window.

        - If the ``locator`` is ``MAIN`` (default, case-insensitive),
          the main window is selected.

        - If the ``locator`` is ``CURRENT`` (case-insensitive), nothing is
          done. This effectively just returns the current window handle.

        - If the ``locator`` is not a string, it is expected to be a list
          of window handles _to exclude_. Such a list of excluded windows
          can be get from `Get Window Handles` prior to doing an action that
          opens a new window.

        The ``timeout`` is used to specify how long keyword will poll to select
        the new window. 

        Example:
        | `Click Link`      | popup1      |      | # Open new window |
        | `Select Window`   | example     |      | # Select window using default strategy |
        | `Title Should Be` | Pop-up 1    |      |
        | `Click Button`    | popup2      |      | # Open another window |
        | ${handle} = | `Select Window`   | NEW  | # Select latest opened window |
        | `Title Should Be` | Pop-up 2    |      |
        | `Select Window`   | ${handle}   |      | # Select window using handle |
        | `Title Should Be` | Pop-up 1    |      |
        | `Select Window`   | MAIN        |      | # Select the main window |
        | `Title Should Be` | Main        |      |
        | ${excludes} = | `Get Window Handles` | | # Get list of current windows |
        | `Click Link`      | popup3      |      | # Open one more window |
        | `Select Window`   | ${excludes} |      | # Select window using excludes |
        | `Title Should Be` | Pop-up 3    |      |

        
        """
        epoch = time.time()
        timeout = epoch if is_falsy(timeout) else timestr_to_secs(timeout) + epoch
        try:
            return self.driver.current_window_handle
        except NoSuchWindowException:
            pass
        finally:
            self._window_manager.select(locator, timeout)

    @keyword
    def close_window(self):
        """Closes currently opened pop-up window."""
        self.driver.close()

    @keyword
    def get_window_handles(self):
        """Return all current window handles as a list.

        Can be used as a list of windows to exclude with `Select Window`.

        
        """
        return self.driver.window_handles

    
    @keyword
    def get_window_identifiers(self):
        """Returns and logs id attributes of all known browser windows."""
        ids = [info.id for info in self._window_manager.get_window_infos()]
        return self._log_list(ids)

    @keyword
    def get_window_names(self):
        """Returns and logs names of all known browser windows."""
        names = [info.name for info in self._window_manager.get_window_infos()]
        return self._log_list(names)

    @keyword
    def get_window_titles(self):
        """Returns and logs titles of all known browser windows."""
        titles = [info.title for info in self._window_manager.get_window_infos()]
        return self._log_list(titles)

    @keyword
    def get_locations(self):
        """Returns and logs URLs of all known browser windows."""
        urls = [info.url for info in self._window_manager.get_window_infos()]
        return self._log_list(urls)

    @keyword
    def maximize_browser_window(self):
        """Maximizes current browser window."""
        self.driver.maximize_window()

    @keyword
    def get_window_size(self):
        """Returns current window width and height as integers.

        See also `Set Window Size`.

        Example:
        | ${width} | ${height}= | `Get Window Size` |
        """
        size = self.driver.get_window_size()
        return size['width'], size['height']

    @keyword
    def set_window_size(self, width, height):
        """Sets current windows size to given ``width`` and ``height``.

        Values can be given using strings containing numbers or by using
        actual numbers. See also `Get Window Size`.

        Browsers have a limit how small they can be set. Trying to set them
        smaller will cause the actual size to be bigger than the requested
        size.

        Example:
        | `Set Window Size` | 800 | 600 |
        """
        return self.driver.set_window_size(int(width), int(height))

    @keyword
    def get_window_position(self):
        """Returns current window position.

        Position is relative to the top left corner of the screen. Returned
        values are integers. See also `Set Window Position`.

        Example:
        | ${x} | ${y}= | `Get Window Position` |
        """
        position = self.driver.get_window_position()
        return position['x'], position['y']

    @keyword
    def set_window_position(self, x, y):
        """Sets window position using ``x`` and ``y`` coordinates.

        The position is relative to the top left corner of the screen,
        but some browsers exclude possible task bar set by the operating
        system from the calculation. The actual position may thus be
        different with different browsers.

        Values can be given using strings containing numbers or by using
        actual numbers. See also `Get Window Position`.

        Example:
        | `Set Window Position` | 100 | 200 |
        """
        self.driver.set_window_position(int(x), int(y))

    def _log_list(self, items, what='item'):
        msg = [
            'Altogether %s %s%s.'
            % (len(items), what, plural_or_not(items))
        ]
        for index, item in enumerate(items):
            msg.append('%s: %s' % (index + 1, item))
        self.info('\n'.join(msg))
        return items



from rflint.common import TestRule, KeywordRule, GeneralRule, ERROR, WARNING

import re


#class LineTooLong(GeneralRule):
#    '''Check that a line is not too long (configurable; default=100)'''

#    severity = WARNING
#    maxchars = 100

#    def configure(self, maxchars):
#        self.maxchars = int(maxchars)

#    def apply(self, robot_file):
#        for linenumber, line in enumerate(robot_file.raw_text.split("\n")):
#            if len(line) > self.maxchars:
#                message = "Line is too long (exceeds %s characters)" % self.maxchars
#                self.report(robot_file, message, linenumber+1, self.maxchars)

#class TrailingBlankLines(GeneralRule):
#    '''Check for multiple blank lines at the end of a file

#    This is a configurable. The default value is 2. 
#    '''

#    severity = WARNING
#    max_allowed = 2

#    def configure(self, max_allowed):
#        self.max_allowed=int(max_allowed)

#    def apply(self, robot_file):
        # I realize I'm making two full passes over the data, but
        # python is plenty fast enough. Even processing a file with
        # over six thousand lines, this takes a couple of
        # milliseconds.  Plenty fast enough for the intended use case,
        # since most files should be about two orders of magnitude
        # smaller than that.

#        match=re.search(r'(\s*)$', robot_file.raw_text)
#        if match:
#            count = len(re.findall(r'\n', match.group(0)))
#            if count > self.max_allowed:
#                numlines = len(robot_file.raw_text.split("\n"))
#                message = "Too many trailing blank lines"
#                linenumber = numlines-count
#                self.report(robot_file, message, linenumber+self.max_allowed, 0)


#class TrailingWhitespace(GeneralRule):
#    severity = WARNING

#    def apply(self, robot_file):
#        for linenumber, line in enumerate(robot_file.raw_text.split("\n")):
#            if len(line) != len(line.rstrip()):
#                message = "Line has trailing whitespace"
#                self.report(robot_file, message, linenumber+1)


#class FileTooLong(GeneralRule):
#    '''Verify the file has fewer lines than a given threshold.

#    You can configure the maximum number of lines. The default is 300.
#    '''

#    severity = WARNING
#    max_allowed = 300

#    def configure(self, max_allowed):
#        self.max_allowed = int(max_allowed)

#    def apply(self, robot_file):
#        lines = robot_file.raw_text.split("\n")
#        if len(lines) > self.max_allowed:
#            message = "File has too many lines (%s)" % len(lines)
#            linenumber = self.max_allowed+1
#            self.report(robot_file, message, linenumber, 0)



# -*- coding: utf-8 -*-

#  Copyright 2019-  DNB
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

GENSON_GENERATED_KEYWORDS = ("$schema", "anyOf", "items", "patternProperties",
                             "properties", "required", "type")

SCHEMA_KEYWORDS = {
    "common": {
        "draft-04": ("enum", "type", "allOf", "anyOf", "oneOf", "not",
                   "definitions"),
        "draft-06": ("enum", "type", "allOf", "anyOf", "oneOf", "not",
                   "definitions", "const", "contains", "propertyNames"),
        "draft-07": ("enum", "type", "allOf", "anyOf", "oneOf", "not",
                   "definitions", "const", "contains", "propertyNames")
    },
    "integer": {
        "draft-04": ("multipleOf", "maximum", "exclusiveMaximum", "minimum",
                   "exclusiveMinimum"),
        "draft-06": ("multipleOf", "maximum", "exclusiveMaximum", "minimum",
                   "exclusiveMinimum"),
        "draft-07": ("multipleOf", "maximum", "exclusiveMaximum", "minimum",
                   "exclusiveMinimum")
    },
    "number": {
        "draft-04": ("multipleOf", "maximum", "exclusiveMaximum", "minimum",
                   "exclusiveMinimum"),
        "draft-06": ("multipleOf", "maximum", "exclusiveMaximum", "minimum",
                   "exclusiveMinimum"),
        "draft-07": ("multipleOf", "maximum", "exclusiveMaximum", "minimum",
                   "exclusiveMinimum")
    },
    "string": {
        "draft-04": ("format", "maxLength", "minLength", "pattern"),
        "draft-06": ("format", "maxLength", "minLength", "pattern"),
        "draft-07": ("format", "maxLength", "minLength", "pattern")
    },
    "object": {
        "draft-04": ("maxProperties", "minProperties", "required",
                   "additionalProperties", "properties",
                   "patternProperties", "dependencies"),
        "draft-06": ("maxProperties", "minProperties", "required",
                   "additionalProperties", "properties",
                   "patternProperties", "dependencies"),
        "draft-07": ("maxProperties", "minProperties", "required",
                   "additionalProperties", "properties",
                   "patternProperties", "dependencies")
    },
    "array": {
        "draft-04": ("additionalItems", "items", "maxItems", "minItems",
                   "uniqueItems"),
        "draft-06": ("additionalItems", "items", "maxItems", "minItems",
                   "uniqueItems"),
        "draft-07": ("additionalItems", "items", "maxItems", "minItems",
                   "uniqueItems")
    }
}
